"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc
import typing


class Metric:
    """Abstract class that any metric is to inherit from"""

    @abc.abstractmethod
    def __add__(self, rhs: typing.Any):
        ...

    @abc.abstractmethod
    def __truediv__(self, rhs: typing.Any):
        ...


class TerminalMetric(abc.ABC, Metric):  # Allows for enhanced type checking
    """Abstract class that a terminal metric is to inherit from.

    A Terminal metric is one that is not contains pure data.
    Examples: Real, String, etc
    """


class NonTerminalMetric(abc.ABC, Metric):  # Allows for enhanced type checking
    """Abstract class that a nonterminal metric is to inherit from.

    A NonTerminal metric is a metric that itself contains other metrics.
    Example: Vector of Reals
    """



"""
This module defines backend entities that maintain the state of a platform and defines the dynamics model used to
calculate state transitions between time steps, given applied controls. In this example, the dynamics model is
a 1D Double Integrator.
"""

import typing

import numpy as np
import scipy.integrate
import scipy.spatial
from pydantic import BaseModel, ConfigDict


class Deputy1dValidator(BaseModel):
    """
    This module validates that Deputy1D configs contain a name and initial state values.
    """

    name: str
    x: typing.Any = None
    xdot: typing.Any = None
    model_config = ConfigDict(arbitrary_types_allowed=True)


class Deputy1D:  # noqa: PLW1641
    """
    1D point mass spacecraft with a +/- thruster and Double Integrator dynamics

    States
        x
        x_dot

    Controls
        thrust
            range = [-1, 1] Newtons

    Parameters
    ----------
    m: float
        Mass of spacecraft in kilograms, by default 12
    integration_method: str
        Numerical integration method passed to dynamics model.
    kwargs:
        Additional keyword arguments passed to Deputy1dValidator
    """

    def __init__(self, m=12, integration_method="RK45", **kwargs):
        self.config = self._get_config_validator()(**kwargs)
        self.name = self.config.name
        self.dynamics = Docking1dDynamics(m=m, integration_method=integration_method)

        self.control_default = np.zeros((1,))
        self.control_min = -1.0
        self.control_max = 1.0

        self._state = self._build_state()
        self.state_dot = np.zeros_like(self._state)

    def __eq__(self, other):
        if isinstance(other, Deputy1D):
            eq = (self.velocity == other.velocity).all()
            return eq and (self.position == other.position).all()
        return False

    @classmethod
    def _get_config_validator(cls):
        return Deputy1dValidator

    def step(self, step_size, action=None):
        """
        Executes a state transition simulation step for the entity

        Parameters
        ----------
        step_size : float
            duration of simulation step in seconds
        action : np.ndarray, optional
            Control action taken by entity, by default None resulting in a control of control_default.

        Raises
        ------
        KeyError
            Raised when action dict key not found in control map
        ValueError
            Raised when action is not one of the required types
        """

        if action is None:
            control = self.control_default.copy()
        elif isinstance(action, np.ndarray):
            control = action.copy()
        else:
            raise ValueError("action must be type np.ndarray")

        # enforce control bounds
        control = np.clip(control, self.control_min, self.control_max)

        # compute new state if dynamics were applied
        self.state, self.state_dot = self.dynamics.step(step_size, self.state, control)

    def _build_state(self):
        # builds initial state?
        return np.array([self.config.x.m, self.config.xdot.m], dtype=np.float32)

    @property
    def state(self) -> np.ndarray:
        """
        Returns copy of entity's state vector

        Returns
        -------
        np.ndarray
            copy of state vector
        """
        return self._state.copy()

    @state.setter
    def state(self, value: np.ndarray):
        self._state = value.copy()

    @property
    def position(self):
        """
        get 1d position vector
        """
        position = np.zeros(1)
        position[0] = self._state[0].copy()
        return position

    @property
    def velocity(self):
        """
        get 1d velocity vector
        """
        velocity = np.zeros(1)
        velocity[0] = self._state[1].copy()
        return velocity


class Docking1dDynamics:
    """
    State transition implementation for generic Linear Ordinary Differential Equation dynamics models of the form
    dx/dt = Ax+Bu. Computes next state through numerical integration of differential equation.

    Parameters
    ----------
    state_min : float or np.ndarray
        Minimum allowable value for the next state. State values that exceed this are clipped.
        When a float, represents single limit applied to entire state vector.
        When an ndarray, each element represents the limit to the corresponding state vector element.
    state_max : float or np.ndarray
        Maximum allowable value for the next state. State values that exceed this are clipped.
        When a float, represents single limit applied to entire state vector.
        When an ndarray, each element represents the limit to the corresponding state vector element.
    angle_wrap_centers: np.ndarray
        Enables circular wrapping of angles. Defines the center of circular wrap such that angles are within [center+pi, center-pi].
        When None, no angle wrapping applied.
        When ndarray, each element defines the angle wrap center of the corresponding state element.
        Wrapping not applied when element is NaN.
    integration_method : string
        Numerical integration method used by dynamics solver. One of ['RK45', 'Euler'].
        'RK45' is slow but very accurate.
        'Euler' is fast but very inaccurate.
    """

    def __init__(
        self,
        state_min: float | np.ndarray = -np.inf,
        state_max: float | np.ndarray = np.inf,
        angle_wrap_centers: np.ndarray | None = None,
        m: float = 12.0,
        integration_method: str = "RK45",
    ):
        self.state_min = state_min
        self.state_max = state_max
        self.angle_wrap_centers = angle_wrap_centers
        self.m = m
        self.A, self.B = self._gen_dynamics_matrices()
        self.integration_method = integration_method

    def step(self, step_size: float, state: np.ndarray, control: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """
        Computes the dynamics state transition from the current state and control input

        Parameters
        ----------
        step_size : float
            Duration of the simation step in seconds.
        state : np.ndarray
            Current state of the system at the beginning of the simulation step.
        control : np.ndarray
            Control vector of the dynamics model.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            tuple of the systems's next state and the state's instantaneous time derivative at the end of the step
        """

        if self.integration_method == "RK45":
            sol = scipy.integrate.solve_ivp(self.compute_state_dot, (0, step_size), state, args=(control,))

            next_state = sol.y[:, -1]  # save last timestep of integration solution
            state_dot = self.compute_state_dot(step_size, next_state, control)
        else:
            raise ValueError(f"invalid integration method '{self.integration_method}'")

        next_state = np.clip(next_state, self.state_min, self.state_max)
        next_state = self._wrap_angles(next_state)
        return next_state, state_dot

    def _wrap_angles(self, state):
        wrapped_state = state.copy()
        if self.angle_wrap_centers is not None:
            wrap_idxs = np.logical_not(np.isnan(self.angle_wrap_centers))

            wrapped_state[wrap_idxs] = ((wrapped_state[wrap_idxs] + np.pi) % (2 * np.pi)) - np.pi + self.angle_wrap_centers[wrap_idxs]

        return wrapped_state

    def compute_state_dot(self, t: float, state: np.ndarray, control: np.ndarray) -> np.ndarray:
        """
        Computes the instataneous time derivative of the state vector

        Parameters
        ----------
        t : float
            Time in seconds since the beginning of the simulation step.
            Note, this is NOT the total simulation time but the time within the individual step.
        state : np.ndarray
            Current state vector at time t.
        control : np.ndarray
            Control vector.

        Returns
        -------
        np.ndarray
            Instantaneous time derivative of the state vector.
        """
        state_dot = np.matmul(self.A, state) + np.matmul(self.B, control)

        # clip state_dot by state limits
        lower_bounded_states = state <= self.state_min
        upper_bounded_state = state >= self.state_max

        state_dot[lower_bounded_states] = np.clip(state_dot[lower_bounded_states], 0, np.inf)
        state_dot[upper_bounded_state] = np.clip(state_dot[upper_bounded_state], -np.inf, 0)

        return state_dot

    def _gen_dynamics_matrices(self):
        """
        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The A and B matrix defining 1D Double Integrator dynamics
        """

        m = self.m

        A = np.array(
            [[0, 1], [0, 0]],
            dtype=np.float32,
        )

        B = np.array(
            [[0], [1 / m]],
            dtype=np.float32,
        )

        return A, B



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Define the abstract TestCaseManager and TestCaseStrategy classes, which provide the context and strategy for
running evaluation episodes with a variety of EPPs.

Author: John McCarroll
"""

import abc
from typing import Any

import pandas as pd
from pydantic import BaseModel

TestCaseIndex = int


class TestCaseStrategyValidator(BaseModel):
    """
    Validator for TestCaseStrategy classes.
    """

    # config: dict = {}


class TestCaseStrategy(abc.ABC):
    """
    Abstract class to define template for TestCaseStrategy.
    This interface handles EPP specific test case management and rllib config mutation.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.config: TestCaseStrategyValidator = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[TestCaseStrategyValidator]:
        """
        Method to return validator for TestCaseStrategy
        """
        return TestCaseStrategyValidator

    @abc.abstractmethod
    def update_rllib_config(self, config: dict) -> dict:
        """
        This method is responsible for mutating the RLLibConfig, overriding or wrapping the EPPs defined in the task configs.

        Parameters
        ----------
        config : dict
            The task's rllib config

        Returns
        -------
        Tconfig : dict
            The task's rllib config
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_test_cases(self) -> pd.DataFrame | list[dict[str, Any]]:
        """
        This method is responsible for fetching the data structure representing configured/scheduled test cases.
        This should NOT return an empty list or empty dataframe. The canonical 'empty' or default test case is [{}]
        which indicates that there are no changes to the default epp parameters.

        Returns
        -------
        test_cases
            The evaluation's test cases
        """
        raise NotImplementedError

    def get_num_test_cases(self) -> int:
        """
        This method is responsible for returning the number of test cases to run. There should be at minimum
        1 test case.
        NOTE: This is not necessarially the same as len(self.test_cases).

        Returns
        -------
        num_test_cases
            The number of evaluation episodes to run (ie. the number of test cases)
        """
        num_test_cases = 0

        test_cases = self.get_test_cases()

        if isinstance(test_cases, pd.DataFrame):
            num_test_cases = len(test_cases.index)

        elif isinstance(test_cases, list):
            num_test_cases = len(test_cases)

        else:
            raise RuntimeError(f"Invalid test_cases {test_cases}\nExpected [pd.DataFrame, list], got type '{type(test_cases)}'")

        if num_test_cases <= 0:
            raise RuntimeError("Too few test cases")

        return num_test_cases

    @abc.abstractmethod
    def get_test_case_index(self, episode_id: int) -> TestCaseIndex:
        """Gets the index of the test case for the specified episode_id"""
        raise NotImplementedError

    def get_test_case(self, index: TestCaseIndex) -> pd.Series | dict[str, Any]:
        return self.get_test_cases()[index]


class NoTestCases(TestCaseStrategy):
    def update_rllib_config(self, config: dict) -> dict:  # noqa: PLR6301
        return config

    def get_test_cases(self) -> pd.DataFrame | list[dict[str, Any]]:  # noqa: PLR6301
        return [{}]

    def get_test_case_index(self, episode_id: int) -> TestCaseIndex:  # noqa: PLR6301
        return 0



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from abc import ABC, abstractmethod


class ConfigUpdate(ABC):
    """Abstract class to update a configuration"""

    @abstractmethod
    def update(self, config: dict) -> dict:
        """Perform update

        Args:
            config (dict): config to update

        Returns:
            dict: updated configuration
        """
        raise RuntimeError("Must Instantiate this method")



"""
This module defines the measurement and control properties for Double Integrator spacecraft sensors and controllers.
"""

from typing import Annotated

from pydantic import Field, StrictFloat

from corl.libraries.property import BoxProp


class PositionProp(BoxProp):
    """
    Position sensor properties.

    name : str
        sensor property name
    low : list[float]
        minimum bounds of sensor output
    high : list[float]
        maximum bounds of sensor output
    unit : str
        unit of measurement for sensor output
    description : str
        description of sensor properties
    """

    name: str = "position"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-150.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [150.0]
    unit: str = "meter"
    description: str = "Position Sensor Properties"


class VelocityProp(BoxProp):
    """
    Velocity sensor properties.

    name : str
        sensor property name
    low : list[float]
        minimum bounds of sensor output
    high : list[float]
        maximum bounds of sensor output
    unit : str
        unit of measurement for sensor output
    description : str
        description of sensor properties
    """

    name: str = "velocity"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-50.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [50.0]
    unit: str = "meter / second"
    description: str = "Velocity Sensor Properties"


class ThrustProp(BoxProp):
    """
    Thrust control properties.

    name : str
        control property name
    low : list[float]
        minimum bounds of control input
    high : list[float]
        maximum bounds of control input
    unit : str
        unit of measurement for control input
    description : str
        description of control properties
    """

    name: str = "thrust"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-1.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [1.0]
    unit: str = "newtons"
    description: str = "Direct Thrust Control"



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Module with base implementations for Observations
"""
from __future__ import annotations

import logging

import numpy as np
from pydantic import BaseModel

from corl.libraries.observation_extractor import ExtractorSet, ObservationExtractor, ObservationSpaceExtractor
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator


class ObservationExtractorValidator(BaseModel):
    """
    fields: The Fields the extractor must access to the nested obs
        ex: [ObserveSensor_Sensor_AltitudeDiff, direct_observation_diff]
    indices: List of indices to extract from the glue
    """

    fields: list[int | str]
    indices: int | list[int] = []

    def construct_extractors(self):
        """
        Builds extractor methods for extracting value, observation_space, and unit from an observation glue

        Parameters
        ----------
        platforms : List[str]
            The platforms the glue is observing, needed to compute the glue's prefix

        Returns
        -------
        ExtractorSet
            Named Tuple of value, space, and unit extractors
        """

        def obs_extractor(obs, *_, full_extraction=False):
            indices = []
            if full_extraction:
                indices = self.indices
            return ObservationExtractor(observation=obs, fields=self.fields, indices=indices)

        def obs_space_extractor(obs, *_):
            return ObservationSpaceExtractor(observation_space=obs, fields=self.fields)

        def unit_extractor(obs, *_):
            return ObservationSpaceExtractor(observation_space=obs, fields=self.fields)

        return ExtractorSet(obs_extractor, obs_space_extractor, unit_extractor)

    @staticmethod
    def get_curr_and_next_observation(extractor, observation, next_observation, allow_array: bool = False):
        """Helper function to extract the current and next observation"""
        curr_metric = extractor(observation)
        next_metric = extractor(next_observation)

        if not isinstance(curr_metric, np.ndarray):
            raise RuntimeError("The extracted observation is not a type that is known how to handle")

        if len(curr_metric) != len(next_metric):
            raise RuntimeError("Length of arrays do not match, this is a nonop")
        if not allow_array:
            if len(curr_metric) != 1:
                raise RuntimeError("The observation attempting to do potential based shaping is not a scalar, unsure how to proceed")
            curr_metric = curr_metric[0]
            next_metric = next_metric[0]
        return (curr_metric, next_metric)


class BaseMeasurementOperationValidator(RewardFuncBaseValidator):
    """
    observation: Dict of observation extractor arguments described in ObservationExtractorValidator
    """

    observation: ObservationExtractorValidator


class BaseMeasurementOperation(RewardFuncBase):
    """Base class for any reward that is to operate on a measurement of some kind"""

    @staticmethod
    def get_validator() -> type[BaseMeasurementOperationValidator]:
        return BaseMeasurementOperationValidator

    def __init__(self, **kwargs) -> None:
        self.config: BaseMeasurementOperationValidator
        super().__init__(**kwargs)
        self._logger = logging.getLogger(self.name)
        self.extractor: ExtractorSet = self.config.observation.construct_extractors()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Print a visualization to screen
"""
import io
import os

from tabulate import tabulate

from corl.evaluation import metrics
from corl.evaluation.metrics.alerts import Alert
from corl.evaluation.metrics.types.nonterminals.dict import Dict
from corl.evaluation.metrics.types.nonterminals.timed_value import TimedValue
from corl.evaluation.metrics.types.nonterminals.vector import Vector
from corl.evaluation.metrics.types.terminals.discrete import Discrete
from corl.evaluation.metrics.types.terminals.rate import Rate
from corl.evaluation.metrics.types.terminals.real import Real
from corl.evaluation.metrics.types.terminals.string import String
from corl.evaluation.metrics.types.terminals.void import Void
from corl.evaluation.visualization.visualization import Visualization


class Print(Visualization):
    """Visualizer that prints to stdout"""

    def __init__(
        self,
        event_table_print: bool = False,
        event_table_field_skip: list[str] | None = None,
        *args,
        **kwargs,
    ):
        """Print data as a event table to both terminal and file"""
        super().__init__(*args, **kwargs)

        self.event_table_print = event_table_print

        if event_table_field_skip is None:
            self.event_table_field_skip = []
        else:
            self.event_table_field_skip = event_table_field_skip
        # self.event_table_field_skip = []

    def metric_prints(
        self, name: str, metric: metrics.metric.Metric, multiple_lines: bool = False, indent: str = "  ", indent_level: int = 0
    ) -> str:
        """Formats a metric type to console print"""
        if isinstance(metric, Discrete):
            string = str(metric.value.name)
        elif isinstance(metric, Real):
            string = str(metric.value)
        elif isinstance(metric, String):
            string = metric.value
        elif isinstance(metric, Rate):
            string = f"{metric.occurrence}/{metric.total}"
        elif isinstance(metric, Vector):
            new_line = "\n"
            string = f"{new_line.join([self.metric_prints(name, item) for item in metric.arr])}"
            if len(metric.arr) > 10:
                with open(os.path.join(self.out_folder, f"{name}.csv"), "w", encoding="utf-8") as fp:
                    fp.write(string.replace(" - ", ","))
                string = f"<{fp.name}>"
            else:
                string = string.replace(",", ",\n")
        elif isinstance(metric, TimedValue):
            string = f"{metric.time} - {self.metric_prints(name, metric.value)}"
        elif isinstance(metric, Dict):
            if multiple_lines:
                internal_data = {
                    key: self.metric_prints(
                        name=name, metric=value, multiple_lines=multiple_lines, indent=indent, indent_level=indent_level + 1
                    )
                    for key, value in metric.data.items()
                }
                string = "".join([f"\n{indent*(indent_level+1)}{key}: {value}" for key, value in internal_data.items()])
            else:
                string = f"{metric.data}"
        elif isinstance(metric, Void):
            string = "No Data"
        else:
            raise RuntimeError(f"Unknown metric type {type(metric)}")

        return string

    # I am okay with this method being long
    def visualize(self) -> None:  # noqa: PLR0915
        """Execute visualization"""

        # ###########################
        # Generate Summary

        summary_buf = io.StringIO()

        if len(self._post_processed_data.world.alerts) > 0:
            print("#################################", file=summary_buf)
            print("Alerts", file=summary_buf)
            alert_types: dict[str, list[Alert]] = {}
            for alert in self._post_processed_data.world.alerts:
                if alert.type not in alert_types:
                    alert_types[alert.type] = []
                alert_types[alert.type].append(alert)

            for alert_type, alert_items in alert_types.items():
                print("  ", alert_type, file=summary_buf)
                for item in alert_items:
                    print("    ", item.name, file=summary_buf)

        print("#################################", file=summary_buf)
        print("Metrics:", file=summary_buf)

        # print any meta metrics
        # print("  meta:")
        for evaluation_metric in self._post_processed_data.world.metrics:
            if evaluation_metric in self.event_table_field_skip:
                continue
            print(
                f"    {evaluation_metric}: "
                f"{self.metric_prints(evaluation_metric, self._post_processed_data.world.metrics[evaluation_metric])}",
                file=summary_buf,
            )

        # Iterate over each agent and print
        for policy_id in self._post_processed_data.participants:
            print(f"  {policy_id}:", file=summary_buf)

            for evaluation_metric in self._post_processed_data.participants[policy_id].metrics:
                if evaluation_metric in self.event_table_field_skip:
                    continue
                printed_data = self.metric_prints(
                    evaluation_metric,
                    self._post_processed_data.participants[policy_id].metrics[evaluation_metric],
                    multiple_lines=True,
                    indent="  ",
                    indent_level=2,
                )
                print(f"    {evaluation_metric}: {printed_data}", file=summary_buf)

        print(summary_buf.getvalue())

        evaluation_summary_file = self.out_folder / "evaluation_summary.txt"
        with open(evaluation_summary_file, "w", encoding="utf-8") as fp:
            fp.write(summary_buf.getvalue())

            print(f"Evaluation Summary written to {evaluation_summary_file}")

        # ###########################
        # Generate Event Table

        # Print Event Table
        headers = ["meta"]
        headers.extend(list(self._post_processed_data.participants))

        table_data = []
        for idx, event in enumerate(self._post_processed_data.world.events):
            row_data = []

            sub_table_headers = ["metric", "value"]
            sub_table_data = []
            for metric in self._post_processed_data.world.events[idx].metrics:
                if metric in self.event_table_field_skip:
                    continue
                sub_table_data.append(
                    [metric, self.metric_prints(metric, self._post_processed_data.world.events[idx].metrics[metric], True)]
                )

            row_data.append(tabulate(sub_table_data, sub_table_headers))

            # Put each agents metrics in the table
            for agent in self._post_processed_data.participants:
                cell = []
                sub_table_headers = ["metric", "value"]
                sub_table_data = []
                for metric_name in self._post_processed_data.participants[agent].events[idx].metrics:
                    if metric_name in self.event_table_field_skip:
                        continue

                    metric = self._post_processed_data.participants[agent].events[idx].metrics[metric_name]  # type: ignore

                    if isinstance(metric, metrics.types.nonterminals.vector.Vector):
                        if len(metric.arr) > 10:
                            # TODO clong: this code block is unfixable and needs to be rewritten
                            #         # If everything is timed value then save to CSV
                            #         if all([isinstance(item, metrics.types.nonterminals.timed_value.TimedValue) for item in metric.arr]):

                            #             csv_file_name = self.out_folder / f"test_case_{idx}/{metric_name}.csv"
                            #             with open(csv_file_name, "w", encoding="utf-8") as fp:

                            #                 # check that make sure all units are right

                            #                 # Write units as header to csv
                            #                 assert isinstance(metric.arr[0], metrics.types.nonterminals.timed_value.TimedValue)
                            #                 fp.write(f"{metric.arr[0].time.u[1][0]},{metric.arr[0].value.value.units.value[1][0]}\n")

                            #                 # Write line by line with no units
                            #                 for item in metric.arr:
                            #                     fp.write(f"{item.time.value},{item.value.value.value}\n")
                            #             sub_table_data.append([metric_name, f"<{csv_file_name}>"])
                            #         else:
                            sub_table_data.append([metric_name, "<skipped>"])
                        continue

                    sub_table_data.append(
                        [
                            metric_name,
                            self.metric_prints(
                                metric_name, self._post_processed_data.participants[agent].events[idx].metrics[metric_name], True
                            ),
                        ]
                    )

                    cell.append(
                        f"{metric_name}: {self.metric_prints(metric_name, self._post_processed_data.participants[agent].events[idx].metrics[metric_name], True)}"  # noqa
                    )

                row_data.append(tabulate(sub_table_data, sub_table_headers))

            # Now the header for each entry in the table will be by test_case_number
            table_data.extend([[f"Test_Case_{event._data.test_case}"], row_data])  # noqa: SLF001

            table_str = tabulate(table_data, headers, "fancy_grid")
            table_str = table_str.replace("\n", "\n    ")

        if self.event_table_print:
            print(f"    {table_str}")

        if not self.out_folder.exists():
            self.out_folder.mkdir()

        event_table_file = self.out_folder / "event_table.txt"
        with open(event_table_file, "w", encoding="utf-8") as fp:
            fp.write("    " + table_str)

            print(f"Evaluation Event Table written to {event_table_file}")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from pydantic import BaseModel, Field, ImportString


class FactoryLoader(BaseModel):
    """Object factory which matches jsonargparse interface"""

    class_path: ImportString
    init_args: dict = Field(default_factory=dict)
    config: dict = Field(default_factory=dict)

    @classmethod
    def resolve_factory(cls, value, info):
        if isinstance(value, dict):
            factory = FactoryLoader(**value)
            if issubclass(factory.class_path, BaseModel):
                return factory.class_path.model_validate({**factory.init_args, **factory.config}, context=info.context)
            return factory.class_path(**factory.init_args, **factory.config)

        return value



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import typing

from corl.agents.base_agent import AgentParseInfo, BaseAgent, PlatformParseInfo
from corl.episode_parameter_providers import EpisodeParameterProvider
from corl.libraries.plugin_library import PluginLibrary
from corl.simulators.base_parts import BasePlatformPart


def create_agent_sim_configs(
    agents: dict[str, AgentParseInfo],
    agent_platforms: dict[str, PlatformParseInfo],
    sim_class: typing.Callable,
    epp_registry: dict[str, EpisodeParameterProvider],
    *,
    multiple_workers: bool = False,
) -> tuple[dict[str, BaseAgent], dict]:
    """
    Creates  dictionary of agent configs used byt the simmulator

    Parameters
    ----------
    agents: dictionary of agent names and configs
    sim_class: simulator class used by environment
    avail_platforms: available platforms in the simulator

    Returns
    -------
    None
    """
    sim_agent_configs = {}
    agent_dict = {}
    platform_part_dict: dict[str, list[BasePlatformPart]] = {}
    platform_cfg_dict: dict[str, list[dict]] = {}

    for platform_name in agent_platforms:
        platform_part_dict[platform_name] = []
        platform_cfg_dict[platform_name] = []

    for agent_name, agent_configs in agents.items():
        agent_class = agent_configs.class_config.agent(
            **agent_configs.class_config.config,
            epp=epp_registry[agent_name],
            agent_name=agent_name,
            platform_names=agent_configs.platform_names,
            multiple_workers=multiple_workers,
        )
        agent_dict[agent_name] = agent_class

        agent_platforms_dict = {
            platform_name: PluginLibrary.get_platform_from_sim_and_config(
                sim_class=sim_class, config=agent_platforms[platform_name]  # type: ignore[arg-type]
            )
            for platform_name in agent_configs.platform_names
        }
        partial_agent_part_dict = agent_class.get_platform_parts(sim_class, agent_platforms_dict)

        for platform_name in agent_configs.platform_names:
            for cls, cfg in partial_agent_part_dict[platform_name]:
                for i, part_cls in enumerate(platform_part_dict[platform_name]):
                    if cls == part_cls and platform_cfg_dict[platform_name][i] == cfg:
                        break
                else:
                    platform_part_dict[platform_name].append(cls)
                    platform_cfg_dict[platform_name].append(cfg)

    for platform_name, platform_config in agent_platforms.items():
        sim_agent_configs[platform_name] = {
            "platform_config": platform_config,
            "parts_list": list(zip(platform_part_dict[platform_name], platform_cfg_dict[platform_name])),
        }

    return agent_dict, sim_agent_configs



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import dataclasses

from corl.evaluation import metrics
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.processors import Evaluation
from corl.evaluation.metrics.scenario_alert_generators import ScenarioAlertGenerators
from corl.evaluation.metrics.scenario_metric_generators import ScenarioMetricGenerators


@dataclasses.dataclass
class SceneProcessors:
    """Contains the processors for a scene.

    Here "Scene" refers to processors in the world and processors attached to participants
    """

    world: Evaluation
    participants: dict[str, Evaluation]

    @staticmethod
    def from_event(
        event_data: EpisodeArtifact,
        scenario_metric_generators: ScenarioMetricGenerators,
        scenario_alert_generators: ScenarioAlertGenerators,
    ) -> "SceneProcessors":
        """Post Process an Event

        This method is not implemented
        """
        raise NotImplementedError()

    @staticmethod
    def from_evaluation(
        evaluation_data: list[EpisodeArtifact],
        scenario_metric_generators: ScenarioMetricGenerators,
        scenario_alert_generators: ScenarioAlertGenerators,
    ) -> "SceneProcessors":
        """Post process and evaluation

        Args:
            evaluation_data (typing.List[EpisodeArtifact]): Data from evaluation
            scenario_metric_generators (ScenarioMetricGenerators): metric generators
            scenario_alert_generators (ScenarioAlertGenerators): alert generators

        Returns:
            _type_: _description_
        """

        contained_participants: list[str] = []
        for episode_artifact in evaluation_data:
            for agent_id in episode_artifact.steps[0].agents:
                if agent_id not in contained_participants:
                    contained_participants.append(agent_id)

        # Process any world metrics
        world_evaluation_metrics: Evaluation = metrics.processors.Evaluation.process(
            evaluation_data,
            scenario_metric_generators.world,
            scenario_alert_generators.world,
        )

        # Processes each agent's metrics
        participant_evaluation_metrics: dict[str, Evaluation] = {}
        for agent_id in contained_participants:
            # Find the generator to use
            if agent_id in scenario_metric_generators.specific_participant:
                generator = scenario_metric_generators.specific_participant[agent_id]
            elif scenario_metric_generators.default_participant is not None:
                generator = scenario_metric_generators.default_participant
            else:
                raise RuntimeError(f"A generator could not be established to generate metrics for {agent_id}")

            # Find the generator to use
            if agent_id in scenario_metric_generators.specific_participant:
                if agent_id not in scenario_alert_generators.specific_participant:
                    scenario_alert_generators.specific_participant[agent_id] = []

                alert_generator = scenario_alert_generators.specific_participant[agent_id]
            elif scenario_metric_generators.default_participant is not None:
                alert_generator = scenario_alert_generators.default_participant
            else:
                raise RuntimeError(f"A generator could not be established to generate metrics for {agent_id}")

            # process the metrics
            participant_evaluation_metrics[agent_id] = metrics.processors.Evaluation.process(
                evaluation_data,
                generator,
                alert_generator,
                agent_id=agent_id,
            )

        return SceneProcessors(world_evaluation_metrics, participant_evaluation_metrics)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
policy mapping functions
"""
from abc import abstractmethod

from pydantic import BaseModel


class PolicyMappingValidatorBase(BaseModel):
    ...


class PolicyMappingBase:
    """Base class for policy mapping implementations"""

    def __init__(self, **kwargs) -> None:
        self.config = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[PolicyMappingValidatorBase]:
        """
        get validator for this Done Functor
        """
        return PolicyMappingValidatorBase

    @abstractmethod
    def __call__(self, agent_id, episode, worker):
        """
        policy mapping functions
        """
        raise NotImplementedError


class PolicyIsAgent(PolicyMappingBase):
    """Map all agents to same policy"""

    def __call__(self, agent_id, episode, worker):
        return agent_id


class SinglePolicyValidator(PolicyMappingValidatorBase):
    """SinglePolicyValidator"""

    policy_id: str = "blue0"


class SinglePolicy(PolicyMappingBase):
    """Map all agents to same policy"""

    def __init__(self, **kwargs) -> None:
        self.config: SinglePolicyValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[SinglePolicyValidator]:
        """
        get validator for this Done Functor
        """
        return SinglePolicyValidator

    def __call__(self, agentid, episode, worker):
        return self.config.policy_id



from corl.agents.base_agent import NonTrainableBaseAgent


class NoOpAgent(NonTrainableBaseAgent):
    """
    Agent that does not send actions
    """

    def create_action(self, observation):  # noqa: PLR6301
        return {}



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Sensors for GymnasiumSimulator
"""
# mypy really does not like the dynamic properties and dies on them,
# this file must be excluded from linting
# mypy: ignore-errors


import numpy as np
from numpy_ringbuffer import RingBuffer

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import BoxProp, RepeatedProp
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_parts import BaseSensor
from corl.simulators.gymnasium.gymnasium_available_platforms import GymnasiumAvailablePlatformTypeMain
from corl.simulators.gymnasium.gymnasium_simulator import GymnasiumSimulator


class GymnasiumSensorProp(BoxProp):
    name: str = "GymnasiumStateSensor"
    unit: str = "dimensionless"
    description: str = "Gymnasium Space"


def create_gym_state_sensor_prop(obs_space) -> type[GymnasiumSensorProp]:
    class TmpGymnasiumSensorProp(GymnasiumSensorProp):  # type: ignore
        low: list[float] = obs_space.low.tolist()
        high: list[float] = obs_space.high.tolist()

    return TmpGymnasiumSensorProp


class GymnasiumStateSensor(BaseSensor):
    """
    Sensor that reports the observation of a given platform gymnasium environment
    """

    def __init__(self, parent_platform, config=None) -> None:
        obs_space = parent_platform.observation_space

        state_sensor_prop = create_gym_state_sensor_prop(obs_space=obs_space)

        super().__init__(parent_platform=parent_platform, config=config, property_class=state_sensor_prop)

    def _calculate_measurement(self, state) -> Quantity:
        return corl_get_ureg().Quantity(state.obs[self.parent_platform.name].astype("float32"), "dimensionless")


PluginLibrary.AddClassToGroup(
    GymnasiumStateSensor, "Sensor_State", {"simulator": GymnasiumSimulator, "platform_type": GymnasiumAvailablePlatformTypeMain}
)


class GymnasiumSensorRepeatedProp(RepeatedProp):
    """
    GymnasiumSensorProp can be updated via config and valdidate by pydantic
    """

    name: str = "GymnasiumStateRepeatedSensor"
    max_len: int = 10
    # child_space: dict[str, GymnasiumSensorProp]
    description: str = "Gymnasium Space Repeated"


def create_gym_state_repeated_sensor_prop(child_prop) -> type[GymnasiumSensorRepeatedProp]:
    class tmpGymnasiumSensorRepeatedProp(GymnasiumSensorRepeatedProp):
        child_space: dict[str, GymnasiumSensorProp] = {"GymnasiumState": child_prop()}

    return tmpGymnasiumSensorRepeatedProp


class GymnasiumRepeatedStateSensor(BaseSensor):
    """
    Sensor that reports the observation of a given platform gymnasium environment
    """

    def __init__(self, parent_platform, config=None) -> None:
        obs_space = parent_platform.observation_space

        state_sensor_prop = create_gym_state_sensor_prop(obs_space=obs_space)

        repeated_state_sensor_prop = create_gym_state_repeated_sensor_prop(state_sensor_prop)

        self.measurement_properties: RepeatedProp
        super().__init__(config=config, parent_platform=parent_platform, property_class=repeated_state_sensor_prop)

        self.obs_buffer = RingBuffer(capacity=self.measurement_properties.max_len, dtype=np.ndarray)

    def _calculate_measurement(self, state):
        self.obs_buffer.append(corl_get_ureg().Quantity(state.obs[self.parent_platform.name].astype("float32"), "dimensionless"))
        ret = ({"GymnasiumState": x} for x in self.obs_buffer)
        return list(ret)


PluginLibrary.AddClassToGroup(
    GymnasiumRepeatedStateSensor,
    "Sensor_RepeatedState",
    {"simulator": GymnasiumSimulator, "platform_type": GymnasiumAvailablePlatformTypeMain},
)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Extra callbacks that can be for evaluating during training to
generate episode artifacts.
"""
import typing

from ray.rllib.utils.typing import MultiAgentDict

import corl.experiments.rllib_utils.wrappers as rllib_wrappers
from corl.agents.base_agent import BaseAgent
from corl.environment.base_multi_agent_env import BaseCorlMultiAgentEnv
from corl.environment.multi_agent_env import ACT3MultiAgentEnvValidator
from corl.episode_parameter_providers import EpisodeParameterProvider


class CorlEnvWrapperValidator(ACT3MultiAgentEnvValidator):
    """Validation model for the inputs of CorlGroupedAgentEnv"""

    wrapped_env: str = "CorlMultiAgentEnv"


class CorlEnvWrapper(BaseCorlMultiAgentEnv):
    """Environment that wraps another CoRL environment."""

    def __init__(self, config: dict):
        """Initialize CorlEnvWrapper."""
        self.config: CorlEnvWrapperValidator
        super().__init__(config)
        self._wrapped_env: BaseCorlMultiAgentEnv = rllib_wrappers.get_rllib_environment_creator(self.config.wrapped_env)(config)
        self._agent_dict = self._wrapped_env.agent_dict
        self._observation_space = self._wrapped_env.observation_space
        self._action_space = self._wrapped_env.action_space
        self._simulator = self._wrapped_env.simulator
        self._state = self._wrapped_env.state

    @staticmethod
    def get_validator() -> type[CorlEnvWrapperValidator]:
        """Get the validator for this class."""
        return CorlEnvWrapperValidator

    def post_process_trajectory(self, agent_id, batch, episode, policy):
        """easy accessor for calling post process trajectory
            correctly

        Arguments:
            agent_id: agent id
            batch: post processed Batch - be careful modifying
        """
        raise NotImplementedError

    @property
    def local_variable_store(self) -> dict[str, typing.Any]:
        """
        get the local variable store

        Returns
        -------
        Dict
            the local variable store episode id
        """
        return self._wrapped_env.local_variable_store

    @property
    def epp(self) -> EpisodeParameterProvider:
        """
        get the Episode Parameter Provider

        Returns
        -------
        EpisodeParameterProvider
        """
        return self._wrapped_env.epp

    def reset(self, *, seed=None, options=None) -> tuple[MultiAgentDict, MultiAgentDict]:
        """Reset env"""
        return_data = self._wrapped_env.reset(seed=seed, options=options)
        self._update_properties()
        return return_data

    def step(self, action_dict: dict):
        """Reset env"""
        step_return = self._wrapped_env.step(action_dict=action_dict)
        self._update_properties()
        return step_return

    def _update_properties(self):
        self._actions = self._wrapped_env.actions
        self._observation = self._wrapped_env.observation
        self._info = self._wrapped_env.info
        self._done_info = self._wrapped_env.done_info
        self._reward_info = self._wrapped_env.reward_info
        self._episode_id = self._wrapped_env.episode_id
        self._state = self._wrapped_env.state


class GroupedAgentsEnvValidator(CorlEnvWrapperValidator):
    """Validation model for the inputs of GroupedAgentsEnv"""

    groups: dict[str, list]


class GroupedAgentsEnv(CorlEnvWrapper):
    """Wrapping environment that combines groups of agents into single agents"""

    def __init__(self, config: dict):
        """Initialize GroupedAgentsEnv."""
        self.config: GroupedAgentsEnvValidator
        super().__init__(config)

        self._groups = self.config.groups
        self._agent_id_to_group = self._create_agent_id_map(self._groups)
        self._action_space = self._group_items(self._wrapped_env.action_space)
        self._observation_space = self._group_items(self._wrapped_env.observation_space)
        self._info = self._group_items(self._wrapped_env.info)
        self._agent_dict = self._create_agent_dict()

    @staticmethod
    def get_validator() -> type[GroupedAgentsEnvValidator]:
        """Get the validator for this class."""
        return GroupedAgentsEnvValidator

    def reset(self, *, seed=None, options=None) -> tuple[MultiAgentDict, MultiAgentDict]:
        """Reset env"""
        obs, trainable_info = self._wrapped_env.reset(seed=seed, options=options)
        self._observation = self._group_items(obs)
        self.actions.clear()

        return self.observation, self._group_items(trainable_info)

    def step(self, action_dict: dict):
        """Step env"""
        obs, rewards, dones, truncated_dones, info = self._wrapped_env.step(action_dict=self._ungroup_items(action_dict))

        training_obs = self._group_items(obs)
        grouped_rewards = self._group_items(rewards, agg_fn=self._ave)
        grouped_dones = self._group_items(dones, agg_fn=self._all)
        grouped_truncated_dones = self._group_items(truncated_dones, agg_fn=self._all)
        grouped_trainable_info = self._group_items(info)

        if not self._wrapped_env.config.disable_action_history:
            self.actions.append(action_dict)
        self._update_properties()

        return training_obs, grouped_rewards, grouped_dones, grouped_truncated_dones, grouped_trainable_info

    def post_process_trajectory(self, agent_id, batch, episode, policy):
        """easy accessor for calling post process trajectory
            correctly

        Arguments:
            agent_id: agent id
            batch: post processed Batch - be careful modifying
        """

    def _create_agent_dict(self) -> dict[str, BaseAgent]:
        """
        get the agent dict

        Returns
        -------
        Dict
            agents
        """
        grouped_agent_dict = self._group_items(self._wrapped_env.agent_dict)
        for group in list(grouped_agent_dict.keys()):
            for agent_name, agent in grouped_agent_dict[group].items():
                grouped_agent_dict[f"{group}/{agent_name}"] = agent
            grouped_agent_dict.pop(group)

        return grouped_agent_dict

    def _update_properties(self):
        self._observation = self._group_items(self._wrapped_env.observation)
        self._info = self._group_items(self._wrapped_env.info)
        self._done_info = self._group_items(self._wrapped_env.done_info)
        self._reward_info = self._group_items(self._wrapped_env.reward_info)
        self._episode_id = self._wrapped_env.episode_id
        self._state = self._wrapped_env.state

    def _ungroup_items(self, grouped_dict):
        ungrouped_dict = type(grouped_dict)()
        for group_name, agents in self._groups.items():
            for agent in agents:
                ungrouped_dict[agent] = grouped_dict[group_name][agent]
        return ungrouped_dict

    @staticmethod
    def _create_agent_id_map(groups):
        agent_id_to_group = {}
        for group_id, agent_ids in groups.items():
            for agent_id in agent_ids:
                if agent_id in agent_id_to_group:
                    raise ValueError(f"Agent id {agent_id} is in multiple groups")
                agent_id_to_group[agent_id] = group_id
        return agent_id_to_group

    def _group_items(self, ungrouped_items, agg_fn=lambda x: x):
        grouped_items = type(ungrouped_items)()
        for agent_id, item in ungrouped_items.items():
            if agent_id in self._agent_id_to_group:
                group_id = self._agent_id_to_group[agent_id]
                if group_id in grouped_items:
                    continue  # already added

                group_out = type(ungrouped_items)()
                for agent in self._groups[group_id]:
                    group_out[agent] = ungrouped_items[agent]

                grouped_items[group_id] = agg_fn(group_out)
            else:
                grouped_items[agent_id] = item
        return grouped_items

    @staticmethod
    def _ave(grouped_items: dict):
        """returns average of value in dict"""
        values = grouped_items.values()
        return sum(values) / len(values)

    @staticmethod
    def _all(grouped_items: dict):
        """returns true if all values in dict are true otherwise false"""
        return all(grouped_items.values())



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import contextlib
import functools
import typing

import numpy as np
import ray
from gymnasium.utils import seeding
from numpy.random import Generator, RandomState
from pydantic import field_validator

from corl.episode_parameter_providers import EpisodeParameterProvider, EpisodeParameterProviderValidator, ParameterModel, Randomness
from corl.libraries.factory import Factory


class RayActorProxy:
    """Creates a 'proxy' to a named ray actor"""

    def __init__(self, actor_id: str, namespace: str | None = None):
        self._actor_id = actor_id
        self._namespace = namespace

    @functools.cached_property
    def actor(self) -> ray.actor.ActorHandle:
        """Retrieves the named ray actor"""
        if not ray.is_initialized():
            raise RuntimeError("Cannot get named actor without ray")
        return ray.get_actor(self._actor_id, self._namespace)


class RemoteEpisodeParameterProviderValidator(EpisodeParameterProviderValidator):
    """Validation model for the inputs of RemoteEpisodeParameterProvider"""

    internal_class: type[EpisodeParameterProvider]
    internal_config: dict[str, typing.Any] = {}
    actor_name: str
    namespace: str | None = None

    @field_validator("internal_class")
    @classmethod
    def internal_not_remote(cls, v):
        """Confirm that internal class is not also remote"""
        assert v != RemoteEpisodeParameterProvider
        return v


class RemoteEpisodeParameterProvider(EpisodeParameterProvider):
    """Wrap EpisodeParameterProvider as a ray remote actor and manage data passing between ray processes."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.config = typing.cast(RemoteEpisodeParameterProviderValidator, self.config)

        ray.remote(self.config.internal_class).options(  # type: ignore
            name=self.config.actor_name,
            namespace=self.config.namespace,
            lifetime="detached",
        ).remote(
            **self.config.internal_config, **kwargs
        )  # I could not get this to format

        self._rap = RayActorProxy(self.config.actor_name, namespace=self.config.namespace)

    @staticmethod
    def get_validator() -> type[RemoteEpisodeParameterProviderValidator]:
        return RemoteEpisodeParameterProviderValidator

    @staticmethod
    def wrap_epp_factory(epp_factory: Factory, actor_name: str, namespace: str | None = None) -> Factory:
        """Wraps an existing EpisodeParameterProvider Factory as a RemoteEpisodeParameterProvider"""
        if not issubclass(epp_factory.type, EpisodeParameterProvider):
            raise TypeError(f"Invalid Factory.type: {epp_factory.type}, {EpisodeParameterProvider.__qualname__} required")

        remote_config = {
            "namespace": namespace,
            "actor_name": actor_name,
            "internal_class": epp_factory.type,
            "internal_config": epp_factory.config,
        }

        return Factory(type=RemoteEpisodeParameterProvider, config=remote_config)

    def kill_actor(self) -> None:
        """Kill the underlying actor used by this provider."""
        with contextlib.suppress(ray.exceptions.RaySystemError):
            ray.kill(self._rap.actor)

    def reset(self):
        ray.get(self._rap.actor.reset.remote())

    def shutdown(self):
        # self.kill_actor()
        ...

    def _do_get_params(self, rng: Randomness, env_epp_ctx: dict | None) -> tuple[ParameterModel, int | None, dict | None]:
        if isinstance(rng, Generator):
            seed = rng.integers(low=0, high=1000000)
            new_rng = np.random.default_rng(seed=seed)
        elif isinstance(rng, RandomState):
            seed = rng.randint(low=0, high=1000000)
            new_rng, _ = seeding.np_random(seed)
        else:
            raise RuntimeError(f"rng type provided to function was {rng}, but this class only knows numpy Generator or RandomState")

        return ray.get(self._rap.actor.get_params.remote(new_rng, env_epp_ctx))

    def compute_metrics(self) -> dict[str, typing.Any]:
        return ray.get(self._rap.actor.compute_metrics.remote())

    def update(self, results: dict, rng: Randomness) -> None:
        if isinstance(rng, Generator):
            seed = rng.integers(low=0, high=1000000)
            new_rng = np.random.default_rng(seed=seed)
        elif isinstance(rng, RandomState):
            seed = rng.randint(low=0, high=1000000)
            new_rng, _ = seeding.np_random(seed)
        else:
            raise RuntimeError(f"rng type provided to function was {rng}, but this class only knows numpy Generator or RandomState")

        ray.get(self._rap.actor.update.remote(results, new_rng))

    def save_checkpoint(self, checkpoint_path) -> None:
        ray.get(self._rap.actor.save_checkpoint.remote(checkpoint_path))

    def load_checkpoint(self, checkpoint_path) -> None:
        ray.get(self._rap.actor.load_checkpoint.remote(checkpoint_path))



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This module contains logic to serialize different platforms
"""


from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer
from corl.simulators.docking_1d.properties import PositionProp as Docking1dPos
from corl.visualization.platform_plotting_deserializer import PlatformDeserializerPlotlyAnimation


class serialize_Docking_1d(PlatformSerializer):
    """
    This is a implementation of the abstract class PlatformSerializer.
    This provides functionality for serializing a 1D docking platform.
    """

    def serialize(self, platform):  # noqa: PLR6301
        """
        Parameters
        ----------
        platform
            The incoming 1d docking platform object that needs to be serialized

        Returns
        -------
        dict
            the serialized representation of the 1d docking platform.
        """
        dictionary = {
            "name": platform.name,
            "position": platform.position,
            "velocity": platform.velocity,
            "sim_time": platform.sim_time,
            "operable": platform.operable,
        }

        dictionary["controllers"] = {}
        for controller in platform.controllers.values():
            dictionary["controllers"][controller.name] = {"applied_controls": controller.get_applied_control()}

        dictionary["sensors"] = {}
        for sensor in platform.sensors.values():
            dictionary["sensors"][sensor.name] = {"measurement": sensor.get_measurement()}

        return dictionary


class Serialize_Pong(PlatformSerializer):
    """
    This is a implementation of the abstract class PlatformSerializer.
    This provides functionality for serializing a Pong Platform to be stored
    in the episode artifact.
    """

    def serialize(self, platform):  # noqa: PLR6301
        """
        Parameters
        ----------
        platform
            The incoming pong platform object that needs to be serialized

        Returns
        -------
        dict
            the serialized representation of the pong platform.
        """
        serialized_data = {
            "name": platform.name,
            "ball_hits": platform.paddle.ball_hits,
            "position": (platform.paddle.x, platform.paddle.y),
            "paddle_height": platform.paddle.height,
            "paddle_width": platform.paddle.width,
            "operable": platform.operable,
            "paddle_type": platform.paddle_type.name,
        }

        serialized_data["controllers"] = {}
        for controller in platform.controllers.values():
            serialized_data["controllers"][controller.name] = {"applied_controls": controller.get_applied_control()}

        serialized_data["sensors"] = {}
        for sensor in platform.sensors.values():
            serialized_data["sensors"][sensor.name] = {"measurement": sensor.get_measurement()}

        return serialized_data


class DeserializeDocking1dPlotlyAnimation(PlatformDeserializerPlotlyAnimation):
    """
    The deserializer for Docking-1d. This is an example
    of how to return the necessary minimal data for
    rendering a plotly trajectory animation within the streamlit application.
    """

    def get_position(self, serialized_data) -> tuple | list:  # noqa: PLR6301
        return serialized_data["position"]

    def get_position_units(self, serialized_data: dict) -> tuple | list:  # noqa: PLR6301
        position_units = (Docking1dPos().unit for _ in serialized_data["position"])
        return tuple(position_units)

    def get_platform_name(self, serialized_data: dict) -> str:  # noqa: PLR6301
        return serialized_data["name"]

    def get_dimensionality(self, serialized_data: dict) -> int:  # noqa: PLR6301
        return len(serialized_data["position"])

    def get_metadata_for_text_display(self, serialized_data) -> dict[str, int | float | str]:  # noqa: PLR6301
        return {"sim_time": serialized_data["sim_time"], "velocity": serialized_data["velocity"].item()}


class DeserializePongPlotlyAnimation(PlatformDeserializerPlotlyAnimation):
    """
    The deserializer for Pong. This is an example
    of how to return the necessary minimal data for
    rendering a plotly trajectory animation within the streamlit application.
    """

    def get_position(self, serialized_data) -> tuple | list:  # noqa: PLR6301
        """Gets the position from the serialized data"""
        return serialized_data["position"]

    def get_position_units(self, serialized_data: dict) -> tuple | list:  # noqa: PLR6301
        """Returns the position units - for Pong, the position units are the coordinates
        of the display screen and are unit-less."""
        position_units = ("None" for _ in serialized_data["position"])
        return tuple(position_units)

    def get_platform_name(self, serialized_data) -> str:  # noqa: PLR6301
        """Returns the name of the platform"""
        return serialized_data["name"]

    def get_dimensionality(self, serialized_data: dict) -> int:  # noqa: PLR6301
        """Returns the dimensionality of the world"""
        return len(serialized_data["position"])

    def get_paddle_dimensions(self, serialized_data: dict):  # noqa: PLR6301
        return {"width": serialized_data["paddle_width"], "height": serialized_data["paddle_height"]}

    def get_metadata_for_text_display(self, serialized_data):  # noqa: PLR6301
        """Returns data in a flattened dictionary that can be optionally displayed
        above each platform"""
        return {
            "ball_hits": serialized_data["ball_hits"],
        }



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Base animator class for streamlit application.
"""
from abc import ABC, abstractmethod
from typing import TypeVar

import matplotlib.colors as mcolors
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from pydantic import BaseModel, ConfigDict

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.libraries.functor import Functor


class FrameAxisBounds(BaseModel):
    """The frame axis bounds for the animation (min, max)"""

    xbounds: tuple[float | int, float | int]
    ybounds: tuple[float | int, float | int]
    zbounds: tuple[float | int, float | int] | None = None


class Polygon(BaseModel):
    """A shape is defined by the name of the shape and points that
    draw the shape. The z-coordinate should be set to None when the shape is
    2-Dimensional."""

    name: str
    x_coordinates: np.ndarray | (list | tuple)
    y_coordinates: np.ndarray | (list | tuple)
    z_coordinates: np.ndarray | (list | tuple) | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseAnimatorValidator(BaseModel):
    """Validator for the BaseAnimator"""

    episode_artifact: TypeVar(EpisodeArtifact)  # type: ignore
    platform_serializer_map: dict[str, str]
    platform_color_map: dict[str, str] | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseAnimator(ABC):
    """Abstract class for creating a streamlit animator."""

    def __init__(
        self, episode_artifact: EpisodeArtifact, platform_serializer_map: dict[str, str], platform_color_map: dict[str, str] | None = None
    ) -> None:
        """
        Parameters
        ----------
        episode_artifact : EpisodeArtifact
            An episode artifact.
        """
        self.config = self.get_validator()(
            episode_artifact=episode_artifact, platform_color_map=platform_color_map, platform_serializer_map=platform_serializer_map
        )
        self.get_platform_color_map_from_platform_name()

    @staticmethod
    def get_validator() -> type[BaseAnimatorValidator]:
        """Get the validator

        Returns
        -------
        Type[BaseAnimatorValidator]
            Validator for BaseAnimator
        """
        return BaseAnimatorValidator

    def get_platform_color_map_from_platform_name(self):
        """Helper function to creates a map that between platform names and a CSS color.
        When the platform name is already in the colormap, skips it.
        """
        platform_names = list(self.config.episode_artifact.platform_to_agents.keys())
        platform_color_map = self.config.platform_color_map
        if platform_color_map is None:
            platform_color_map = {}
        color_map = mcolors.CSS4_COLORS
        for name in platform_names:
            if name in platform_color_map:
                continue
            # Checks if a color is part of the platform name
            possible_color = [color for color_name, color in color_map.items() if color_name in name]
            if possible_color:
                platform_color_map[name] = possible_color[0]
        self.config.platform_color_map = platform_color_map

    def get_deserializer(self):
        """
        Returns the deserializer from the serializer found in the episode artifact.
        Grabs the deserializer from the platform serializer map
        """
        module = self.config.episode_artifact.platform_serializer.__module__
        cls_name = self.config.episode_artifact.platform_serializer.__class__.__name__
        class_name = f"{module}.{cls_name}"
        deserializer = self.config.platform_serializer_map.get(class_name)
        if deserializer is None:
            return None
        return Functor(functor=deserializer).functor

    @abstractmethod
    def get_dimensionality(self) -> int:
        """Returns the dimensionality of the animation

        Returns
        -------
        int
            Dimensionality of the animation world (e.g. 2 or 3)
        """
        raise NotImplementedError

    @abstractmethod
    def set_frame_axis_bounds(self, frame_axis_bounds: FrameAxisBounds):
        """Abstract method for setting the axis bounds"""
        raise NotImplementedError

    @abstractmethod
    def get_metadata_names(self) -> list[str]:
        """Abstract method that returns a list of metadata values that could be
        used to select additional information to plot above the platform.
        Return and empty list to exclude this feature"""

    @abstractmethod
    def generate_animation(self, selected_metadata: list[str] | None = None) -> tuple[go.Figure, list[go.Figure], dict[str, pd.DataFrame]]:
        """Abstract method. Main method to generate the animation"""



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import enum


class Scopes(enum.IntEnum):
    """Scopes which a metric can be computed on"""

    EVENT = enum.auto()
    """An Event is a single flight
    """

    EVALUATION = enum.auto()
    """An Evaluation is a list of events
    """

    TOURNAMENT = enum.auto()
    """A Tournament is a list of evaluations
    """


def from_string(name: str) -> Scopes:
    """Generate a scope enumeration from string"""
    if name == "event":
        return Scopes.EVENT
    if name == "evaluation":
        return Scopes.EVALUATION
    if name == "tournament":
        return Scopes.TOURNAMENT
    raise RuntimeError(f"Unknown metric scope: {name}")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import dataclasses
import typing

from corl.evaluation.metrics.metric import TerminalMetric

Enum = typing.TypeVar("Enum")


@dataclasses.dataclass
class Discrete(typing.Generic[Enum], TerminalMetric):  # noqa: PLW1641
    """Terminal Metric to contain a value of an enumeration"""

    value: Enum

    def __add__(self, rhs: typing.Any):
        raise RuntimeError("Can not add to a discrete metric")

    def __eq__(self, rhs: typing.Any):
        return self.value == rhs

    def __truediv__(self, denominator: typing.Any):
        raise RuntimeError("Can not divide a Discrete metric")



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Network Explainability Plotter
"""

import typing

import numpy as np
import pandas as pd
import plotly.express as px
from ray.rllib.utils.typing import AgentID

from corl.visualization.network_explainability.network_explainability import PPOTorchNetworkExplainability


class PPOTorchNetworkExplainabilityPlotter(PPOTorchNetworkExplainability):
    """
    Renders plots for various explainability tools.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.color_palette = px.colors.qualitative.Alphabet
        self.height = 600

    def render_agg_feat_importance_grad_based(self, feats_dict: dict[str, typing.Any] | None = None, **kwargs):
        """Plotting wrapper for `get_agg_feat_importance_grad_based`, see method for
        input parameters.

        Parameters
        ----------
        feats_dict : Optional[Dict[str, Any]]
            An optional features dict to pass in. When one is not provided, calls
            `self.get_agg_feat_importance_grad_based(**kwargs), by default None

        Returns
        -------
        plotly.graph_objs.Figure
            Figure object.
        """
        if feats_dict is None:
            _feats_dict: dict[AgentID, dict[str, typing.Any]] = self.get_agg_feat_importance_grad_based(**kwargs)
        else:
            _feats_dict = feats_dict

        dataframe_list = []
        for policy_id, feats in _feats_dict.items():
            # Manipuate the data into a long formatted dataframe for plotting
            dataframe = pd.DataFrame(feats)
            dataframe = dataframe.reset_index()
            dataframe.rename(columns={"index": "input_obs"}, inplace=True)
            dataframe = pd.melt(dataframe, value_name="feat_importance", id_vars=["input_obs"], var_name="policy_output")
            dataframe["policy_id"] = policy_id
            dataframe_list.append(dataframe)
        dataframe_all = pd.concat(dataframe_list, axis=0, ignore_index=True)

        fig = px.bar(
            data_frame=dataframe_all,
            x="feat_importance",
            y="policy_output",
            color="input_obs",
            facet_col="policy_id",
            orientation="h",
            height=self.height,
            color_discrete_sequence=self.color_palette,
            labels={
                "input_obs": "Policy Input",
                "policy_output": "Network Output",
                "feat_importance": "Average Normalized Feature Importance",
            },
            facet_col_wrap=2,
        )

        fig.update_layout(yaxis={"automargin": True}, font={"size": 16})
        return fig

    def render_value_function_residual_dist(self, vf_residuals: dict[str, typing.Any] | None = None):
        """Plotter to render the distribution of the value function residuals.

        Parameters
        ----------
        vf_residuals : typing.Optional[typing.Dict[str, typing.Any]], optional
            The value function residuals where the key is a reference name (e.g. policy, agent or some other descriptor),
            when vf_residuals is not provided, will call self.vf_residuals to retrieve, by default None

        Returns
        -------
        plotly.graph_objs.Figure
            Figure object.
        """
        dfs = []
        if vf_residuals is None:
            vf_residuals = self.vf_residuals
        for name, residuals_list in vf_residuals.items():
            one_agent_df = pd.DataFrame(np.concatenate(residuals_list), columns=["vf_residuals"])
            one_agent_df["policy_id"] = name
            dfs.append(one_agent_df)
        dataframe = pd.concat(dfs, axis=0, ignore_index=True)
        return px.histogram(
            dataframe,
            x="vf_residuals",
            color="policy_id",
            color_discrete_sequence=self.color_palette,
            marginal="box",
            height=self.height,
            labels={"policy_id": "Policy", "vf_residuals": "Value Targets - Value Predictions"},
        )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Module with base implementations for Observations
"""
from collections import defaultdict

import flatten_dict
import numpy as np
from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch


class RandomActionPolicy(Policy):
    """Random action policy."""

    def __init__(self, observation_space, action_space, config):
        Policy.__init__(self, observation_space, action_space, config)
        self.view_requirements = {key: value for key, value in self.view_requirements.items() if key != SampleBatch.PREV_ACTIONS}

    def compute_actions(
        self,
        obs_batch,
        state_batches=None,
        prev_action_batch=None,
        prev_reward_batch=None,
        info_batch=None,
        episodes=None,
        explore=None,
        timestep=None,
        **kwargs,
    ):
        if not isinstance(obs_batch, list):
            obs_batch = [obs_batch]

        actions_batch = defaultdict(list)
        for _ in obs_batch:
            random_actions = self.action_space.sample()
            for key, value in flatten_dict.flatten(random_actions).items():
                actions_batch[key].append(value)

        for key, value in actions_batch.items():
            actions_batch[key] = np.array(value)

        return flatten_dict.unflatten(actions_batch), [], {}

    def learn_on_batch(self, samples):  # noqa: PLR6301
        return {}

    def get_weights(self):  # noqa: PLR6301
        return {}

    def set_weights(self, weights):  # noqa: PLR6301
        return



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
EnvironmentDefaultCallbacks
"""

import contextlib
import warnings
from collections import defaultdict

import numpy as np
from flatten_dict import flatten
from gymnasium.utils import seeding
from ray.rllib import BaseEnv
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.typing import AgentID, PolicyID

from corl.dones.done_func_base import DoneStatusCodes

SHORT_EPISODE_THRESHOLD = 5


def log_done_status(env, episode):
    """
    Log done status codes to done_status/{platform}/{status}
    """

    def process_done_entry(platform, data, prefix=""):
        if not data:
            return
        at_least_table = [0] * len(DoneStatusCodes)
        episode_codes = set(data.values())

        for indx, status in enumerate(DoneStatusCodes):
            metric_key = f"{prefix}done_status/{platform}/{status}"
            if status in episode_codes:
                episode.custom_metrics[metric_key] = 1
                for x in range(indx, len(at_least_table)):
                    at_least_table[x] += 1
            else:
                episode.custom_metrics[metric_key] = 0

        for indx, status in enumerate(DoneStatusCodes):
            if indx in [0, 4]:
                continue
            metric_key = f"{prefix}done_status/{platform}/AT_LEAST_{status}"
            episode.custom_metrics[metric_key] = at_least_table[indx]

    if env.state.episode_state:
        for platform, data in env.state.episode_state.items():
            process_done_entry(platform=platform, data=data)

    if env.state.agent_episode_state:
        for agent, data in env.state.agent_episode_state.items():
            process_done_entry(platform=agent, data=data, prefix="agent_")


def log_done_info(env, episode):
    """
    Log done info to done_results/{platform}/{done_name}
    """
    if env.done_info:
        for platform_id, platform_data in env.done_info.items():
            for done_name, done_value in platform_data.items():
                episode.custom_metrics[f"done_results/{platform_id}/{done_name}"] = int(done_value)
            episode.custom_metrics[f"done_results/{platform_id}/NoDone"] = int(not any(platform_data.values()))


class EnvironmentDefaultCallbacks(DefaultCallbacks):
    """
    This is the default class for callbacks to be use in the Environment class.
    To make your own custom callbacks set the EnvironmentCallbacks in your derived class
    to be a class that subclasses this class (EnvironmentDefaultCallbacks)

    Make sure you call the super function for all derived functions or else there will be unexpected callback behavior
    """

    DEFAULT_METRIC_OPS = {
        "min": np.min,
        "max": np.max,
        "median": np.median,
        "mean": np.mean,
        "var": np.var,
        "std": np.std,
        "sum": np.sum,
        "nonzero": np.count_nonzero,
    }

    def on_episode_start(
        self,  # noqa: PLR6301
        *,
        worker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy],
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        """Callback run on the rollout worker before each episode starts.

        Args:
            worker: Reference to the current rollout worker.
            base_env: BaseEnv running the episode. The underlying
                sub environment objects can be retrieved by calling
                `base_env.get_sub_environments()`.
            policies: Mapping of policy id to policy objects. In single
                agent mode there will only be a single "default" policy.
            episode: EpisodeV2 object which contains the episode's
                state. You can use the `episode.user_data` dict to store
                temporary data, and `episode.custom_metrics` to store custom
                metrics for the episode.
            kwargs: Forward compatibility placeholder.
        """
        super().on_episode_start(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

        episode.user_data["rewards_accumulator"] = defaultdict(float)  # default dict with default value of 0.0

        env = base_env.get_sub_environments()[episode.env_id]
        env.simulator.callbacks.on_episode_start(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

    def on_episode_step(
        self,  # noqa: PLR6301
        *,
        worker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy] | None = None,
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        """Runs on each episode step.

        Args:
            worker: Reference to the current rollout worker.
            base_env: BaseEnv running the episode. The underlying
                sub environment objects can be retrieved by calling
                `base_env.get_sub_environments()`.
            policies: Mapping of policy id to policy objects.
                In single agent mode there will only be a single
                "default_policy".
            episode: EpisodeV2 object which contains episode
                state. You can use the `episode.user_data` dict to store
                temporary data, and `episode.custom_metrics` to store custom
                metrics for the episode.
            kwargs: Forward compatibility placeholder.
        """

        super().on_episode_step(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

        env = base_env.get_sub_environments()[episode.env_id]

        if env.reward_info:
            rewards_accumulator = episode.user_data["rewards_accumulator"]
            for reward_name, reward_val in flatten(env.reward_info, reducer="path").items():
                key = f"rewards_cumulative/{reward_name}"
                rewards_accumulator[key] += reward_val

        env.simulator.callbacks.on_episode_step(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

    def on_episode_end(self, *, worker, base_env: BaseEnv, policies: dict[PolicyID, Policy], episode, **kwargs) -> None:  # noqa: PLR6301
        """
        on_episode_end stores the custom metrics in RLLIB. Note this is on a per glue basis.

        1. read the training information for the current episode
        2. For each metric in each platform interface in each environment
           update metric container

        Parameters
        ----------
        worker: RolloutWorker
            Reference to the current rollout worker.
        base_env: BaseEnv
            BaseEnv running the episode. The underlying
            env object can be gotten by calling base_env.get_sub_environments().
        policies: dict
            Mapping of policy id to policy objects. In single
            agent mode there will only be a single "default" policy.
        episode: MultiAgentEpisode
            EpisodeV2 object which contains episode
            state. You can use the `episode.user_data` dict to store
            temporary data, and `episode.custom_metrics` to store custom
            metrics for the episode.
        """
        if isinstance(episode, Exception):
            return

        # Issue a warning if the episode is short
        # This, should get outputting in the log and may help identify any setup issues
        if episode.length < SHORT_EPISODE_THRESHOLD:
            msg = f"Episode {episode.episode_id!s} length {episode.length} is less than warn threshold {SHORT_EPISODE_THRESHOLD!s}"
            if "params" in episode.user_data:
                msg += "\nparams:\n"
                for key in episode.user_data["params"]:
                    msg += f"  {key}: {episode.user_data['params'][key]!s}"
                    msg += "\n"
            else:
                msg += "\nParams not provided in episode user_data"
            warnings.warn(msg)
        for i in range(1, 11):
            length_bound = SHORT_EPISODE_THRESHOLD * i
            episode.custom_metrics[f"episode_length/{length_bound}"] = episode.length < length_bound

        env = base_env.get_sub_environments()[episode.env_id]
        if env.glue_info:
            for glue_name, metric_val in flatten(env.glue_info, reducer="path").items():
                episode.custom_metrics[glue_name] = metric_val

        if env.reward_info:
            for reward_name, reward_val in flatten(env.reward_info, reducer="path").items():
                key = f"rewards/{reward_name}"
                episode.custom_metrics[key] = reward_val

        log_done_info(env, episode)

        log_done_status(env, episode)

        # Variables
        for key, value in flatten(env.local_variable_store, reducer="path").items():
            with contextlib.suppress(ValueError):
                if isinstance(value, str):
                    continue
                episode.custom_metrics[f"variable/env/{key}"] = float(value.m)
        for agent_name, agent_data in env.agent_dict.items():
            for key, value in flatten(agent_data.local_variable_store, reducer="path").items():
                with contextlib.suppress(ValueError):
                    if isinstance(value, str):
                        continue
                    episode.custom_metrics[f"variable/{agent_name}/{key}"] = float(value.m)
        # Episode Parameter Providers
        metrics = env.epp.compute_metrics()
        for k, v in metrics.items():
            episode.custom_metrics[f"adr/env/{k}"] = v

        for agent_name, agent_data in env.agent_dict.items():
            metrics = agent_data.config.epp.compute_metrics()
            for k, v in metrics.items():
                episode.custom_metrics[f"adr/{agent_name}/{k}"] = v

        # Cumulative Rewards
        for key, value in episode.user_data["rewards_accumulator"].items():
            episode.custom_metrics[key] = value

        env.simulator.callbacks.on_episode_end(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

    def on_postprocess_trajectory(
        self,  # noqa: PLR6301
        *,
        worker,
        episode,
        agent_id: AgentID,
        policy_id: PolicyID,
        policies: dict[PolicyID, Policy],
        postprocessed_batch: SampleBatch,
        original_batches: dict[AgentID, tuple[Policy, SampleBatch]],
        **kwargs,
    ) -> None:
        """
        Called immediately after a policy's postprocess_fn is called.

        You can use this callback to do additional postprocessing for a policy,
        including looking at the trajectory data of other agents in multi-agent
        settings.

        Parameters
        ----------
        worker: RolloutWorker
            Reference to the current rollout worker.
        episode: MultiAgentEpisode
            EpisodeV2 object.
        agent_id: str
            Id of the current agent.
        policy_id: str
            Id of the current policy for the agent.
        policies: dict
            Mapping of policy id to policy objects. In single
            agent mode there will only be a single "default" policy.
        postprocessed_batch: SampleBatch
            The postprocessed sample batch
            for this agent. You can mutate this object to apply your own
            trajectory postprocessing.
        original_batches: dict
            Mapping of agents to their unpostprocessed
            trajectory data. You should not mutate this object.
        """
        super().on_postprocess_trajectory(
            worker=worker,
            episode=episode,
            agent_id=agent_id,
            policy_id=policy_id,
            policies=policies,
            postprocessed_batch=postprocessed_batch,
            original_batches=original_batches,
        )
        episode.worker.foreach_env(lambda env: env.post_process_trajectory(agent_id, postprocessed_batch, episode, policies[policy_id]))

    def on_train_result(self, *, algorithm, result: dict, **kwargs) -> None:  # noqa: PLR6301
        """
        Called at the end of Trainable.train().

        Parameters
        ----------
        algorithm: algorithm
            Current algorithm instance.
        result: dict
            Dict of results returned from algorithm.train() call.
            You can mutate this object to add additional metrics.
        """
        rng, _ = seeding.np_random(seed=algorithm.iteration)

        for epp in result["config"]["env_config"].get("epp_registry", {}).values():
            epp.update(result, rng)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Custom Policy
"""
from abc import abstractmethod
from collections import defaultdict

import flatten_dict
import gymnasium
import numpy as np
from pydantic import ConfigDict, model_validator
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.spaces.space_utils import normalize_action
from ray.rllib.utils.typing import TensorStructType, TensorType

from corl.policies.base_policy import BasePolicyValidator


class CustomPolicyValidator(BasePolicyValidator):
    """Base validator for the CustomPolicy"""

    agent_id: str | None = None

    act_space: gymnasium.Space
    obs_space: gymnasium.Space
    controllers: list[tuple[str, ...]]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def validate_model(self):
        assert len(self.controllers) == len(set(self.controllers)), "controller definitions must be unique"
        sample_control = flatten_dict.flatten(self.act_space.sample())
        for tuple_key in self.controllers:
            assert tuple_key in sample_control, f"controller {tuple_key} not found in action_space: {list(sample_control.keys())}"
        return self


class CustomPolicy(Policy):
    """Custom base policy."""

    def __init__(self, observation_space, action_space, config) -> None:
        self.validated_config: CustomPolicyValidator = self.get_validator()(act_space=action_space, obs_space=observation_space, **config)

        Policy.__init__(self, observation_space, action_space, config)

        self._agent_id = self.validated_config.agent_id
        self._reset_time = 0

        self._reset()

    @staticmethod
    def get_validator() -> type[CustomPolicyValidator]:
        """
        Get the validator for this experiment class,
        the kwargs sent to the experiment class will
        be validated using this object and add a self.config
        attr to the experiment class
        """
        return CustomPolicyValidator

    def _reset(self):
        """This must be overridden in order to reset the state between runs"""

    def learn_on_batch(self, samples):  # noqa: PLR6301
        return {}

    def get_weights(self):  # noqa: PLR6301
        return {}

    def set_weights(self, weights):  # noqa: PLR6301
        return

    def compute_actions_from_input_dict(
        self,
        input_dict: SampleBatch | dict[str, TensorStructType],
        explore: bool | None = None,
        timestep: int | None = None,
        episodes: list[EpisodeV2] | None = None,
        **kwargs,
    ) -> tuple[TensorType, list[TensorType], dict[str, TensorType]]:
        """Computes actions from collected samples (across multiple-agents).

        Takes an input dict (usually a SampleBatch) as its main data input.
        This allows for using this method in case a more complex input pattern
        (view requirements) is needed, for example when the Model requires the
        last n observations, the last m actions/rewards, or a combination
        of any of these.

        Args:
            input_dict: A SampleBatch or input dict containing the Tensors
                to compute actions. `input_dict` already abides to the
                Policy's as well as the Model's view requirements and can
                thus be passed to the Model as-is.
            explore: Whether to pick an exploitation or exploration
                action (default: None -> use self.config["explore"]).
            timestep: The current (sampling) time step.
            episodes: This provides access to all of the internal episodes'
                state, which may be useful for model-based or multi-agent
                algorithms.

        Keyword Args:
            kwargs: Forward compatibility placeholder.

        Returns:
            actions: Batch of output actions, with shape like
                [BATCH_SIZE, ACTION_SHAPE].
            state_outs: List of RNN state output
                batches, if any, each with shape [BATCH_SIZE, STATE_SIZE].
            info: Dictionary of extra feature batches, if any, with shape like
                {"f1": [BATCH_SIZE, ...], "f2": [BATCH_SIZE, ...]}.
        """
        # Default implementation just passes obs, prev-a/r, and states on to
        # `self.compute_actions()`.
        agent_id_batch = []
        raw_obs_batch = []
        epp_info_batch = []
        sim_time_batch = []
        do_reset = False

        for i in range(input_dict.count):  # type: ignore
            episode = episodes[i]  # type: ignore

            # there is a really nasty bug in the agent id code and it comes out as -1 sometimes
            # so we save a valid one and just cache it so we don't have to worry about it
            agent_index = input_dict[SampleBatch.AGENT_INDEX][i]
            agents = episode.get_agents()
            self._agent_id = agents[agent_index] if 0 <= agent_index < len(agents) else self._agent_id
            if self._agent_id is None:
                raise RuntimeError(f"Unable to determine the agent_id for agent_index: {agent_index}")
            # end hack section

            epp_info = episode.last_info_for(self._agent_id)
            if epp_info:
                sim_time = epp_info["env"]["sim_time"]
            else:
                epp_info = {}
                sim_time = self._reset_time

            if episode.length == 0 or sim_time < self._reset_time:
                self._reset_time = sim_time

            if sim_time <= self._reset_time:
                do_reset = True

            # rllib preprocessor add original space attribute to the space
            raw_obs = gymnasium.spaces.utils.unflatten(self.observation_space.original_space, input_dict["obs"][i])  # type: ignore

            sim_time_batch.append(sim_time)
            agent_id_batch.append(self._agent_id)
            epp_info_batch.append(epp_info)
            raw_obs_batch.append(raw_obs)

        if do_reset:
            self._reset()

        return self.compute_actions(
            input_dict[SampleBatch.OBS],
            [s for k, s in input_dict.items() if k[:9] == "state_in_"],
            prev_action_batch=input_dict.get(SampleBatch.PREV_ACTIONS),
            prev_reward_batch=input_dict.get(SampleBatch.PREV_REWARDS),
            info_batch=input_dict.get(SampleBatch.INFOS),  # type: ignore
            explore=explore,
            timestep=timestep,
            episodes=episodes,
            sim_time_batch=sim_time_batch,
            agent_id_batch=agent_id_batch,
            epp_info_batch=epp_info_batch,
            raw_obs_batch=raw_obs_batch,
            **kwargs,
        )

    def compute_actions(
        self,
        obs_batch: list[TensorStructType] | TensorStructType,
        state_batches: list[TensorType] | None = None,
        prev_action_batch: list[TensorStructType] | TensorStructType = None,
        prev_reward_batch: list[TensorStructType] | TensorStructType = None,
        info_batch: dict[str, list] | None = None,
        episodes: list[EpisodeV2] | None = None,
        explore: bool | None = None,
        timestep: int | None = None,
        **kwargs,
    ) -> tuple[TensorType, list[TensorType], dict[str, TensorType]]:
        num_batches = len(obs_batch)

        sim_time_batch = kwargs.get("sim_time_batch", None)
        agent_id_batch = kwargs.get("agent_id_batch", None)
        epp_info_batch = kwargs.get("epp_info_batch", None)
        raw_obs_batch = kwargs.get("raw_obs_batch", [{} for _ in range(num_batches)])

        actions_batch = defaultdict(list)

        for i in range(num_batches):
            actions = self.custom_compute_actions(
                obs_batch[i],  # type: ignore
                platform_obs=raw_obs_batch[i],
                state=state_batches[i] if state_batches is not None and len(state_batches) > i else None,
                prev_action=prev_action_batch[i] if prev_action_batch is not None and len(prev_action_batch) > i else None,
                prev_reward=prev_reward_batch[i] if prev_reward_batch is not None and len(prev_reward_batch) > i else None,
                info=info_batch[i] if info_batch is not None and len(info_batch) > i else None,  # type: ignore
                explore=explore,
                timestep=timestep,
                sim_time=sim_time_batch[i] if sim_time_batch is not None and len(sim_time_batch) > i else None,
                agent_id=agent_id_batch[i] if agent_id_batch is not None and len(agent_id_batch) > i else None,
                epp_info=epp_info_batch[i] if epp_info_batch is not None and len(epp_info_batch) > i else None,
                episode=episodes[i] if episodes is not None and len(episodes) > i else None,
                **kwargs,
            )

            if self.config.get("normalize_actions"):
                actions = normalize_action(action=actions, action_space_struct=self.validated_config.act_space)

            for key, value in flatten_dict.flatten(actions).items():
                actions_batch[key].append(value)

        for key, value in actions_batch.items():
            actions_batch[key] = np.array(value)  # type: ignore

        # # NOTE: ray/rllib is inconsistent in applying these modifications to the returned actions (between training and the
        # # Algorithm.compute_single_action fn) so we do it explicitly here to ensure that the behavior is consistent between them.
        # actions = convert_to_numpy(actions)
        # if isinstance(actions, list):
        #     actions = np.array(actions)

        return flatten_dict.unflatten(actions_batch), [], {}

    @abstractmethod
    def custom_compute_actions(
        self,
        obs: list[TensorStructType] | TensorStructType,
        platform_obs: dict[str, dict],
        state: list[TensorType] | None = None,
        prev_action: list[TensorStructType] | TensorStructType = None,
        prev_reward: list[TensorStructType] | TensorStructType = None,
        info: dict[str, list] | None = None,
        explore: bool | None = None,
        timestep: int | None = None,
        sim_time: float | None = None,
        agent_id: str | None = None,
        epp_info: dict | None = None,
        episode: EpisodeV2 | None = None,
        **kwargs,
    ) -> TensorType:
        """Computes actions for the current policy.

        Args:
            obs: observations.
            state: List of RNN state input batches, if any.
            prev_action: previous action values.
            prev_reward: previous rewards.
            info: info objects.
            explore: Whether to pick an exploitation or exploration action.
                Set to None (default) for using the value of
                `self.config["explore"]`.
            timestep: The current (sampling) time step.

        Keyword Args:
            kwargs: Forward compatibility placeholder

        Returns:
            actions (TensorType): output actions
        """
        raise NotImplementedError



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Network Explainability Tools
"""

import typing
from warnings import warn

import gymnasium
import numpy as np
import torch
from flatten_dict import flatten_dict
from pydantic import BaseModel, ConfigDict
from ray.rllib.algorithms.ppo.ppo_torch_policy import PPOTorchPolicy
from ray.rllib.evaluation import postprocessing as ray_postprocessing
from ray.rllib.models.modelv2 import restore_original_dimensions
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.typing import AgentID

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.visualization.network_explainability.env_policy_transforms import ACT3MultiAgentEnvToPolicyNetworkInputs, AgentCheckpoint

VALUE_FUNCTION_PRED = "value_pred"
SAMPLE_BASED_FEAT_IMPORTANCE = "sample_based"
GRAD_BASED_FEAT_IMPORTANCE = "gradient_based"


class NetworkExplainabilityValidator(BaseModel):
    """Validator for Network Explainability"""

    value_function_prediction: str = VALUE_FUNCTION_PRED
    network_output_per_controller: tuple[str, str] = ("mean", "std")
    ignore_inputs_containing: str = "invalid"
    model_config = ConfigDict(arbitrary_types_allowed=True)


class PPOTorchNetworkExplainability:
    """Generates feature importance metrics and plots."""

    def __init__(
        self,
        agent_checkpoints: list[AgentCheckpoint],
        env_config: dict[str, typing.Any],
        value_function_prediction: str = VALUE_FUNCTION_PRED,
        network_output_controller: tuple[str, str] = ("mean", "std"),
    ) -> None:
        """
        Initializes the policies and environment for a set of trajectories. Each trajectory contained
        in an EpisodeArtifact is expected to be passed into each feature importance method.

        Parameters
        ----------
        agent_checkpoints : typing.List[AgentCheckpoint]
            A list of AgentCheckpoints
        env_config : typing.Union[EnvContext, typing.Dict]
            The environment configuration used to recreate the ACT3MultiAgentEnv
        value_function_prediction : typing.Optional[str], optional
            The name of the value function prediction, by default VALUE_FUNCTION_PRED
        network_output_per_controller : typing.Tuple[str], optional
            This should indicate what the policy networkout outputs. For example, for PPO, the outputs
            are the mean and standard deviation of each controller, by default ('mean', 'std').
        """

        self.config: NetworkExplainabilityValidator = self.get_validator()(
            value_function_prediction=value_function_prediction, network_output_per_controller=network_output_controller
        )

        self.env_policy_input_transformer = ACT3MultiAgentEnvToPolicyNetworkInputs(
            agent_checkpoints=agent_checkpoints, env_config=env_config
        )

        self.feat_importance_gradient_based: dict[AgentID, list[dict[str, typing.Any]]] = self._initialize_feature_dict()
        self.feat_importance_sample_based: dict[AgentID, list[dict[str, typing.Any]]] = self._initialize_feature_dict()
        self.vf_residuals: dict[AgentID, list[typing.Any]] = self._initialize_feature_dict()
        self._network_output_names: list[str] = []

    def _initialize_feature_dict(self) -> dict[AgentID, list[dict[str, typing.Any]]]:
        """Helper function that returns an empty dictionary with the agent-ids as the keys"""

        return {agent_id: [] for agent_id, _ in self.env_policy_input_transformer.policies.items()}

    @staticmethod
    def get_validator() -> type[NetworkExplainabilityValidator]:
        """Gets validator for class"""
        return NetworkExplainabilityValidator

    @property
    def network_output_names(self):
        """A list of strings indicating each network output"""
        return self._network_output_names

    @network_output_names.setter
    def network_output_names(self, value):
        """Custom setter that checks to ensure the value of the outputs have not
        changed between episode artifacts"""
        if self._network_output_names:
            if self._network_output_names != value:
                raise ValueError("Network outputs changed between episoide artifacts.", f" {self._network_output_names} -> {value}")
        else:
            self._network_output_names = value

    def generate_vf_residuals(self, episode_artifact: EpisodeArtifact, agent_id: AgentID | None = None) -> dict[AgentID, np.ndarray]:
        """
        Generates value function residuals for the provided episode,
        and returns the residuals in a dictionary of
        [key: agent_id, values: policy_residuals].

        See https://github.com/ray-project/ray/blob/ray-2.2.0/rllib/evaluation/sampler.py#L756
        for how the postprocessed_batch is generated. The postprocessed_batch, accessible
        in the postprocessed_trajectory method of the default callback, contains the
        value predictions and value targets. This method aims to replicate the logic.

        Parameters
        ----------
        episode_artifact : EpisodeArtifact
            An episode artifact to pull the trajectory from.
        agent_id : typing.Optional[AgentID], optional
            The agent-id when None, the feature importance for all policies will
            be generated, by default None

        Returns
        -------
        typing.Dict[AgentID, np.ndarray]
            A dictionary where the top-level key is the agent id and
            the values are a numpy array of the residuals.
        """

        process_policies: list = list(self.env_policy_input_transformer.policies.keys())
        # If an agent is specified, overwrite the list of policies to process
        if agent_id:
            process_policies = [agent_id]

        vf_residuals_per_policy = {}
        for _agent_id in process_policies:
            # Get policy for current agent
            _policy = self.env_policy_input_transformer.policies[_agent_id]

            # Get a temporary instance of the sample_batch to extract
            # value function predictions
            sample_batch = self.env_policy_input_transformer.raw_observations_to_sample_batch(
                _agent_id, episode_artifact, include_initial_state=True
            )

            # Get value function predictions
            _, _, info = _policy.compute_actions_from_input_dict(sample_batch)
            vf_preds = info["vf_preds"]

            # Get rewards from episode artifact steps: Note this does not include the initial state.
            # These are the rewards received at a given timestep
            rewards = np.array([s.agents[_agent_id].rewards for s in episode_artifact.steps])
            rewards = np.array([np.array(list(reward.values())).sum() for reward in rewards])
            rewards = np.array(rewards)
            # Get dones: does not include the initial state
            # --> rllib SampleBatch expects truncated and terminated conditions not dones
            terminateds = np.zeros_like(rewards)
            truncateds = np.zeros_like(rewards)
            # Check if the episode completed with a done condition.
            # If so, add a True (=1) to the end of the done condition
            # array (assuming here that the done is the last step in
            # the episode).

            _agent_platforms = episode_artifact.agent_to_platforms[_agent_id]
            for _agent_platform in _agent_platforms:
                if any(episode_artifact.dones[_agent_platform].values()):
                    terminateds[-1] = 1
                    truncateds[-1] = 1
                    # If the episode ends in a done condition, remove the value
                    # prediction of the terminal state. It will be replaced by a reward
                    # of 0 when `ray_postprocessing.compute_gae_for_sample_batch` is called
                    vf_preds = vf_preds[:-1]

            # Add rewards, dones, and vf_preds to new sample batch
            # We create a new SampleBatch instance because the
            # previous SampleBatch causes multiple issues due to
            # device-mismatches. This approach is simpler than using
            # the already-defined sample_batch and running the list of
            # required device-swapping functions.
            sample_batch_new = SampleBatch(
                {
                    SampleBatch.REWARDS: rewards,
                    SampleBatch.TERMINATEDS: terminateds,
                    SampleBatch.TRUNCATEDS: truncateds,
                    SampleBatch.VF_PREDS: vf_preds,
                }
            )

            # Get post-processed sample batch from which we can extract
            # the value function targets.
            postprocessed_sample_batch = ray_postprocessing.compute_gae_for_sample_batch(_policy, sample_batch_new)
            vf_targets = postprocessed_sample_batch["value_targets"]
            vf_residuals = vf_targets - vf_preds
            vf_residuals_per_policy[_agent_id] = vf_residuals

        return vf_residuals_per_policy

    @property
    def network_input_names(self):
        """Network input names"""
        return self.env_policy_input_transformer.policy_input_observation_names

    def generate_feat_importance_grad_based(
        self,
        episode_artifact: EpisodeArtifact,
        agent_id: AgentID | None = None,
        policy_network_input_sample_batch_key_map: dict | None = None,
    ) -> dict[AgentID, typing.Any]:
        """Generates gradient-based feature importance for one episode, returns the feature
        importance in a dictionary.

        The features are generated for each observation and output, where the output
        for each policy is shaped: [episode_length, number of observations], where
        the name of each observation is stored in `self.env_policy_transforms.policy_input_observation_names`

        Parameters
        ----------
        episode_artifact : EpisodeArtifact
            An episode artifact to pull the trajectory from.
        agent_id : typing.Optional[AgentID], optional
            The policy-id, when None, the feature importance for all policies will
            be generated, by default None
        policy_network_input_sample_batch_key_map : dict, optional
            A dictionary that maps the policy-id to the expected policy network input SampleBatch key.
            When one is not provided, it assumes the policy network input for all policies is `SampleBatch.OBS`.
            This is needed because in order to generate the feature importance, torch.autograd.grad is manually called.

        Returns
        -------
        typing.Dict[AgentID, typing.Any]
            A dictionary where the top-level key is the policy-id and the values are
            a dictionary of gradient-based feature importances.
        """
        process_policies: list = list(self.env_policy_input_transformer.policies.keys())
        # If an agent is specified, overwrite the list of policies to process
        if agent_id:
            process_policies = [agent_id]
        feat_importance_per_policy = {}

        for _agent_id in process_policies:
            # Policies can have different inputs
            input_key = (
                policy_network_input_sample_batch_key_map.get(_agent_id, SampleBatch.OBS)
                if policy_network_input_sample_batch_key_map
                else SampleBatch.OBS
            )
            sample_batch = self.env_policy_input_transformer.raw_observations_to_sample_batch(_agent_id, episode_artifact)
            gradients_dict = self._get_policy_network_gradients(_agent_id, sample_batch, input_key)
            feat_importance_per_policy[_agent_id] = gradients_dict
        return feat_importance_per_policy

    def append_vf_residuals(self, episode_artifact: EpisodeArtifact) -> None:
        """
        Generates the residuals of the value function for the specified
        episode artifact (for all policies, retrieved as
        `self.policies.keys()`) and appends them to the list of
        residuals for the corresponding policiy_id in
        `self.vf_residuals`.

        Parameters
        ----------
        episode_artifact : EpisodeArtifact
            An episode artifact.

        Returns
        -------
        None
        """

        residuals = self.generate_vf_residuals(episode_artifact)
        for agent_id, residual_values in residuals.items():
            self.vf_residuals[agent_id].append(residual_values)

    def append_feat_importance_gradient_based(
        self, episode_artifact: EpisodeArtifact, policy_network_input_sample_batch_key_map: dict | None = None
    ) -> None:
        """Generates the gradient-based feature importance and
        appends it to `self.feat_importance_gradient_based`.

        Parameters
        ----------
        episode_artifact : EpisodeArtifact
            An episode artifact.
        """
        grad_feat_importance = self.generate_feat_importance_grad_based(
            episode_artifact, policy_network_input_sample_batch_key_map=policy_network_input_sample_batch_key_map
        )
        for agent_id, feat_importance in grad_feat_importance.items():
            self.feat_importance_gradient_based[agent_id].append(feat_importance)

    def get_agg_feat_importance_grad_based(
        self, agent_id: AgentID = None, normalize: bool = True, agg_func: typing.Callable = np.mean, test_case: int | None = None
    ) -> dict[AgentID, dict[str, typing.Any]]:
        """Aggregates the gradient-based feature importance according to the `agg_func` across all observations
        and test cases (defined by `test_case`). For example, when `agg_func=np.mean`, returns a dictionary
        that specifies how important each input is to the output, on average across test cases. Assumes the
        `self.append_feat_importance_gradient_based` has already been called to fill out the array of gradient-based feature importances

        Parameters
        ----------
        agent_id : AgentID, optional
            When provided, only returns the aggregated feature importance for the specified agent. Otherwise, all
            agents are returned, by default None
        normalize : bool, optional
            When True, normalizes the aggregated feature importance so that all features for a given
            output sum up to 1 , by default True
        agg_func : typing.Callable, optional
            Specifies how the features should be aggregated, by default np.mean. The callable must take `axis` as an input.
        test_case : typing.Optional[int], optional
            An integer that represents which test cases to process. When none is specified, processes all
            test cases, by default None. Note the test case is the index of `self.feat_importance_gradient_based`

        Returns
        -------
        typing.Dict[AgentID, typing.Dict[str, typing.Any]]
            A nested dictionary that specifies how important each input is to the output.
        """

        feat_importance = self.feat_importance_gradient_based
        policy_input_observation_names = np.array(self.network_input_names.copy())
        discrete_groups = self.env_policy_input_transformer.discrete_inputs_indicator.copy()
        ignore_columns_idx = self._ignore_policy_input_columns()
        discrete_groups = np.delete(discrete_groups, ignore_columns_idx)
        policy_input_observation_names = np.delete(policy_input_observation_names, ignore_columns_idx)

        # Select policies to process, when None, all policies will be processed.
        policies = [agent_id] if agent_id else list(feat_importance.keys())
        agg_feats: dict[AgentID, dict[str, typing.Any]] = {}

        for _agent_id in policies:
            agg_feats[_agent_id] = {}
            policy_feat_importance = feat_importance[_agent_id]
            assert (
                len(policy_feat_importance) > 0
            ), f"No episode artifacts were processed for {_agent_id}, did you forget to call `append_feat_importance_gradient_based`?"
            agg_feat_per_policy = {}
            # When only one test is specified, grab the test case
            if test_case:
                policy_feat_importance = [policy_feat_importance[test_case]]
            # Iterate over test cases and vertically stack the array of features
            for policy_feat_importance_per_test_case in policy_feat_importance:
                for output_feat_name, feat_array in policy_feat_importance_per_test_case.items():
                    if output_feat_name not in agg_feat_per_policy:
                        agg_feat_per_policy[output_feat_name] = feat_array
                    else:
                        agg_feat_per_policy[output_feat_name] = np.vstack((agg_feat_per_policy[output_feat_name], feat_array))

            # Iterate over the one-hot encoded vector groups and store which columns need to be removed
            unique_groups = np.unique(discrete_groups)
            # 0 is the fill value and does not indicate a group of one-hot encoded values.
            unique_groups = unique_groups[unique_groups.nonzero()]

            remove_columns = np.array([]).astype(int)
            grouped_policy_input_observation_names = policy_input_observation_names.copy()
            for group in unique_groups:
                # The [0] refers to the first element of the tuple. Since
                # discrete_groups is a flattened array, there is only one
                # element in the tuple.
                group_idx = np.where(discrete_groups == group)[0]
                # All one-hot encoded vectors are summed together and stored where the first one-hot encoded
                # vector of the group appears
                # Flag the rest of the one-hot encoded vectors for removal
                remove_columns = np.hstack((remove_columns, group_idx[1:]))
                # Consolidate the one-hot encoded vector names: this takes the first one-hot encoded vector, and replaces the
                # suffix (_0)
                grouped_policy_input_observation_names[group_idx[0]] = policy_input_observation_names[group_idx[0]].replace("_0", "")
            # Updates the input observation names accordingly
            grouped_policy_input_observation_names = np.delete(grouped_policy_input_observation_names, remove_columns)

            # Aggregate the feature importance for each policy, we take the absolute value before the aggregate so that
            # the sign of gradient does not impact the metric (e.g. the direction of the gradient is ignored.)
            for output_feat_name, feat_array in agg_feat_per_policy.items():
                feat_array = np.abs(feat_array)  # noqa: PLW2901
                assert len(discrete_groups) == feat_array.shape[-1]
                # Iterate over the one-hot encoded vectors and aggregate them
                for group in unique_groups:
                    group_idx = np.where(discrete_groups == group)[0]
                    feat_array[:, group_idx[0]] = feat_array[:, group_idx].sum(axis=1)
                feat_array = np.delete(feat_array, remove_columns, axis=1)  # noqa: PLW2901
                # NOTE: Each column represents a different input and each row
                # represents a different observation
                _aggregated = agg_func(feat_array, axis=0)
                # Normalize so that each feature importance adds up to 1
                if normalize:
                    _aggregated = _aggregated / _aggregated.sum()

                assert len(_aggregated) == len(grouped_policy_input_observation_names)
                agg_feats[_agent_id].update({output_feat_name: dict(zip(grouped_policy_input_observation_names, _aggregated))})

        return agg_feats

    def _get_policy_network_gradients(
        self, agent_id: AgentID, sample_batch: SampleBatch, policy_network_input_sample_batch_key: str = SampleBatch.OBS
    ) -> dict[str, np.ndarray]:
        """Returns the gradient of each input with respect to the output feature

        Parameters
        ----------
        agent_id : AgentID
            The agent id.
        sample_batch : SampleBatch
            The inputs to the policy network
        policy_network_input_sample_batch_key : str, optional.
            The key in SampoleBatch that corresponds to the input of the policy network. By default,
            `SampleBatch.OBS`.

        Returns
        -------
        Dict[str, np.ndarray]
            Each array represents the gradient of the input features with respect
            one output feature, where the output feature is the name of the key. For example,
            when the key is the mean of the acceration, the array represents the gradients of all
            observations with respect to the mean acceelration. Each array is shaped: [batch_size, num_input_feats]
        """
        policy_network: PPOTorchPolicy = self.env_policy_input_transformer.policies[agent_id]

        # requires_grad must be set to true in order to compute the gradient below.
        sample_batch[policy_network_input_sample_batch_key].requires_grad = True

        # shape: [batch_size, num_output_feats]
        output_tensor, _ = policy_network.model(sample_batch)
        _, num_output_feats = output_tensor.shape
        value_preds_tensor = policy_network.model.value_function()

        ignore_idx = self._ignore_policy_input_columns()
        expected_num_columns = len(self.network_input_names)

        # The network outputs are ordered accordingly
        for idx in range(num_output_feats):
            # Grabs one output feature (across all samples)
            output_feature = output_tensor[:, idx]
            # Retain the graph so that we can compute gradients wrt all outputs. By default,
            # the graph used to compute the gradient will be freed after the grad call.
            # Shape: [batch_size, num_input_feats]
            grad_tensor = torch.autograd.grad(
                outputs=output_feature.sum(), inputs=sample_batch[policy_network_input_sample_batch_key], retain_graph=True
            )
            # The grad_tensor is a tuple of length 1
            grad_array = grad_tensor[0].detach().numpy()
            # Sanity check on the number of columns
            assert grad_array.shape[-1] == expected_num_columns
            grad_array = self._check_and_process_extra_inputs(
                view_requirements=policy_network.model.view_requirements,
                grad_array=grad_array,
                policy_network_input_sample_batch_key=policy_network_input_sample_batch_key,
            )

            # These columns don't correspond to input features and are used for padding
            if ignore_idx:
                grad_array = np.delete(grad_array, ignore_idx, axis=1)
            # Stack arrays in sequence depth wise (in the third dimension)
            # we can ignore F821 because gradients is defined the first iteration
            gradients: np.ndarray = grad_array if idx == 0 else np.dstack((gradients, grad_array))  # noqa: F821

        assert isinstance(policy_network.action_space, gymnasium.spaces.Dict)
        output_gradients_dict, network_output_names = self._process_outputs(
            action_space=policy_network.action_space, output_features=gradients, num_output_feats=num_output_feats
        )
        # Compute the gradient for the value function output
        grad_tensor = torch.autograd.grad(outputs=value_preds_tensor.sum(), inputs=sample_batch[policy_network_input_sample_batch_key])
        grad_array = grad_tensor[0].detach().numpy()
        grad_array = self._check_and_process_extra_inputs(
            view_requirements=policy_network.model.view_requirements,
            grad_array=grad_array,
            policy_network_input_sample_batch_key=policy_network_input_sample_batch_key,
        )
        if ignore_idx:
            grad_array = np.delete(grad_array, ignore_idx, axis=1)
        output_gradients_dict[self.config.value_function_prediction] = grad_array
        network_output_names.append(self.config.value_function_prediction)
        self.network_output_names = network_output_names
        return output_gradients_dict

    def _check_and_process_extra_inputs(
        self,  # noqa: PLR6301
        view_requirements: dict,
        grad_array: np.ndarray,
        policy_network_input_sample_batch_key: str,
    ):
        """
        When `policy_network_input_sample_batch_key` is not the `SampleBatch.OBS` and the gradient array is 3-dimensional, it's assumed
        the dimensions correspond to [batch_size, sequence_length, num_inputs]. This function (1) checks to ensure that the  policy network
        inputs are derived from `SampleBatch.OBS` and (2) the second dimension of the gradient array corresponds to the length of the
        model view's `shift_arr` and (3) sums the absolute value of the gradient array across the sequence axis.
        """
        if policy_network_input_sample_batch_key != SampleBatch.OBS and len(grad_array.shape) > 2:
            view_reqs = view_requirements[policy_network_input_sample_batch_key]
            assert view_reqs.data_col == SampleBatch.OBS, "Only policy inputs based on the observation are permitted."
            assert grad_array.shape[1] == len(
                view_reqs.shift_arr
            ), """Dimension mismatch:
                the second dimension is assumed to be the the number of observation states that the input depends on, given by the
                length of `ViewRequirements.shift_arr`."""
            grad_array = np.abs(grad_array)
            grad_array = grad_array.sum(axis=1)
            warn(
                f"""The absolute value of the gradient-based feature importance was \
                    aggregated across the {policy_network_input_sample_batch_key} axis. \
                    That is, each feature importance is the |gradient| of the sum of the previous
                    {view_reqs.shift} inputs with respect to each output."""
            )
        return grad_array

    def _process_outputs(
        self, action_space: gymnasium.spaces.Dict, output_features: np.ndarray, num_output_feats: int
    ) -> tuple[dict[str, typing.Any], list[str]]:
        """Helper function to process the output features. Checks the action space type and aggregates
        any Discrete or MultiDiscrete one-hot encoded vectors

        Parameters
        ----------
        action_space : typing.Dict[str, typing.Any]
            Action space
        output_features : np.ndarray
            Output features to check. Expected shape is [batch_size, num_input_features, num_output_features]
        num_output_feats : int
            Number of expected output features

        Returns
        -------
        typing.Dict[str, typing.Any]
            A dictionary of non-aggregated feature importances
        typing.List[str]
            A list of network output names.

        Raises
        ------
        NotImplementedError
            When a Discrete space starts at a value > 0
        NotImplementedError
            For any gymnasium spaces other than MultiDiscrete, Discrete and Box.
        """
        network_output_names = []
        output_feats_dict: dict[str, typing.Any] = {}
        assert isinstance(action_space, gymnasium.spaces.Dict)
        flattened_action_space = flatten_dict.flatten(action_space, reducer="dot")
        # gradients shape: [batch_size, num_input_feats, num_output_feats]
        for action_space_name, _gymnasium_space in flattened_action_space.items():
            # When the gymnasium space is multidiscrete, we aggregate the gradients of the one-hot encoded vectors.
            if isinstance(_gymnasium_space, gymnasium.spaces.MultiDiscrete | gymnasium.spaces.Discrete):
                # This is a mapping of the index to each action space
                ref_idx_dict = restore_original_dimensions(
                    np.expand_dims(np.arange(0, num_output_feats), axis=0), obs_space=action_space, tensorlib=np
                )
                ref_idx_dict = flatten_dict.flatten(ref_idx_dict, reducer="dot")
                # Grabs the index (action space output) corresponding to each action
                idx_range = ref_idx_dict[action_space_name].flatten()
                # Grab the outputs corresponding to this space
                gradients_subset = output_features[:, :, idx_range].copy()

                class_name = _gymnasium_space.__class__.__name__
                start = 0
                if isinstance(_gymnasium_space, gymnasium.spaces.MultiDiscrete):
                    nvecs = _gymnasium_space.nvec
                if isinstance(_gymnasium_space, gymnasium.spaces.Discrete):
                    if _gymnasium_space.start > 0:
                        raise NotImplementedError
                    nvecs = np.array([_gymnasium_space.n])
                else:
                    raise RuntimeError(f"gymnasium_space must be MultiDiscrete or Discrete, got {_gymnasium_space}")
                # The one hot encoding is described here: https://docs.ray.io/en/latest/rllib/rllib-models.html#built-in-preprocessors
                for one_hot_vec_span in nvecs:
                    # sum across the depth dimension
                    # gradients_subset shaped: batch_size, num_input_features, num_output_features]
                    # gradients_agg shaped: [batch_size, num_input_features] - sum is taken across depth
                    gradients_agg = gradients_subset[:, :, start : start + one_hot_vec_span].sum(axis=2)
                    # The output name, e.g. controller_MultiDiscrete[2]
                    output_name = f"{action_space_name}_{class_name}[{one_hot_vec_span}]"
                    network_output_names.append(output_name)
                    output_feats_dict[output_name] = gradients_agg
                    # The start of the next group will be where this group ended
                    start = one_hot_vec_span

            # If its a box, the outputs are the mean and std deviation.
            elif isinstance(_gymnasium_space, gymnasium.spaces.Box):
                _box_network_output_names = self._get_network_output_names(action_space)

                assert output_features.shape[2] == len(_box_network_output_names)
                for _idx, name in enumerate(_box_network_output_names):
                    output_feats_dict[name] = output_features[:, :, _idx]
                    network_output_names.append(name)
            # If its any other type of space, raise a NotImplemetnedError
            else:
                raise NotImplementedError("Only implemented for gymnasium spaces: MultiDiscrete, Discrete and Box.")

        return output_feats_dict, network_output_names

    def _ignore_policy_input_columns(self):
        """Helper function to identify columns that either correspond to a pad column (relevant for repeated obs)
        or a column with the name matching `ignore_inputs_containg` (e.g. invalid)"""
        return [
            idx
            for idx, name in enumerate(self.network_input_names)
            if (name == "") or (self.config.ignore_inputs_containing in name)  # noqa: PLC1901
        ]

    def _filtered_policy_input_columns(self):
        return [
            name
            for name in self.network_input_names
            if (name != "") and (self.config.ignore_inputs_containing not in name)  # noqa: PLC1901
        ]

    def _get_network_output_names(self, action_space_struct: gymnasium.spaces.Dict) -> list[str]:
        """Given a dict of actions, returns a list of network outputs"""
        # Returns a flattened dictionary, where all keys are separated by a dot.
        action_space_struct = flatten_dict.flatten(action_space_struct, reducer="dot")
        action_space_keys = list(action_space_struct.keys())
        network_outputs = []
        for action_name in action_space_keys:
            network_outputs.extend([f"{name}_{action_name}" for name in self.config.network_output_per_controller])
        return network_outputs



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Launch process for performing and visualizing an evaluation
"""
import typing

import jsonargparse

from corl.evaluation.evaluation_artifacts import EvaluationArtifact_Metrics, EvaluationArtifact_Visualization
from corl.evaluation.visualization.visualization import Visualization


def get_args(path=None):
    """
    Obtain the running arguments for the visualization segment of the evaluation framework
    """
    parser = jsonargparse.ArgumentParser(description="visualizer")
    parser.add_argument("--cfg", help="the path to a json/yml file containing the running arguments", action=jsonargparse.ActionConfigFile)

    parser.add_class_arguments(EvaluationArtifact_Metrics, "artifact_metrics", instantiate=True)
    parser.add_class_arguments(EvaluationArtifact_Visualization, "artifact_visualization", instantiate=True)

    # Link location such that if visualization isn't given, the metrics location will be used
    parser.link_arguments("artifact_metrics.location", "artifact_visualization.location", apply_on="instantiate")

    parser.add_argument("--visualizations", type=list[Visualization])

    args = parser.parse_args() if path is None else parser.parse_path(path)
    instantiate = parser.instantiate_classes(args)
    return args, instantiate


def main(instantiate: dict[str, typing.Any] | jsonargparse.Namespace):
    """
    Execute the visualization portion of the evaluation framework

    Parameters:
        args: jsonargparse.Namespace
            configurations for visualization
        instantiated: jsonargparse.Namespace
            instantiated arguments , contains  the actual classes to be run for visualization
    """

    for visual in instantiate["visualizations"]:
        visual.load(instantiate["artifact_metrics"], instantiate["artifact_visualization"])
        visual.visualize()


def pre_main():
    """
    calls gets current args and passes them to main
    """
    _, instantiated = get_args()
    main(instantiated)


if __name__ == "__main__":
    pre_main()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from pydantic import ConfigDict

from corl.glues.base_glue import BaseAgentGlue, BaseAgentGlueValidator


class BaseMultiWrapperGlueValidator(BaseAgentGlueValidator):
    """
    wrapped - the wrapped glue instances
    """

    wrapped: list[BaseAgentGlue]
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseMultiWrapperGlue(BaseAgentGlue):
    """A base object that glues can inherit in order to "wrap" multiple glue instances"""

    def __init__(self, **kwargs) -> None:
        self.config: BaseMultiWrapperGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseMultiWrapperGlueValidator]:
        return BaseMultiWrapperGlueValidator

    def glues(self) -> list[BaseAgentGlue]:
        """Get the wrapped glue instances"""
        return self.config.wrapped

    def set_agent_removed(self, agent_removed: bool = True) -> None:
        super().set_agent_removed(agent_removed)
        for glue in self.glues():
            glue.set_agent_removed()



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Entrypoint for Streamlit App
"""
import jsonargparse
import streamlit as st

from corl.parsers.yaml_loader import load_file
from corl.visualization.streamlit_app.default_layout import DefaultStreamlitPage, Functor


@st.cache_data
def read_config_file(yaml_file: str) -> dict:
    """Returns the parsed configuration file as a dictionary

    Parameters
    ----------
    yaml_file : str
        The path to the yaml file.

    Returns
    -------
    dict:
        The parsed configurations.
    """
    return load_file(yaml_file)


def _parse_display_trajectory_config(config: dict) -> dict:
    """Parses the configs for the trajectory animation section of streamlit"""
    name = DefaultStreamlitPage.display_trajectory_animation_section.__name__
    if name not in config["streamlit_layout"]:
        return config

    config["streamlit_layout"][name]["animator"] = Functor(**config["streamlit_layout"][name]["animator"])
    config["streamlit_layout"][name]["frame_axis_bounds"] = Functor(
        **config["streamlit_layout"][name]["frame_axis_bounds"]
    ).create_functor_object()
    return config


if __name__ == "__main__":
    parser = jsonargparse.ArgumentParser()
    parser.add_argument("--config", help="Path to the yaml file containing the plotly configurations", type=str, default=None)
    args = parser.parse_args()

    st.set_page_config(layout="wide", page_title="Introduction")
    st.session_state.layout_config = {}
    if args.config:
        st.session_state.layout_config = _parse_display_trajectory_config(read_config_file(yaml_file=args.config))

    st.markdown("# Overview")
    st.markdown(
        """
        This application was developed to inspect evaluation trajectories stored in the episode artifacts.
        - The Evaluation From Training page allows the user to inspect episode artifacts generated from a training run
        This implies a logging callback was included as part of training and evaluation was enabled through the `rllib_config.yml`.
        - The Corl Evaluation page allows user to inspect episode artifacts generated from a custom corl evaluation.
        - This streamlit app takes as an optional input, a configuration file that defines the trajectory animation section. See
        `config/streamlit_app/default_streamlit_config.yml` for an example with Pong.
        """
    )
    # NOTE: in order for the argparser to parse correctly and set the `layout_config`
    # this page must fully load before the subsequent pages are loaded.
    st.session_state.main_page_loaded = True



import numpy as np
from plotly.graph_objs import Scatter, Scatter3d
from pygame import Rect

from corl.simulators.pong.pong import Ball
from corl.visualization.plotly_animator import PlotlyTrajectoryAnimation, draw_custom_polygon


class PongAnimation(PlotlyTrajectoryAnimation):
    def _draw_custom_polygons(self, step_idx) -> list[Scatter | Scatter3d] | None:
        polygons = []
        for platform_serialized_data in self.config.episode_artifact.steps[step_idx].platforms:
            agent_name = platform_serialized_data["name"]
            paddle_height = platform_serialized_data["paddle_height"]
            paddle_width = platform_serialized_data["paddle_width"]
            paddle_position = platform_serialized_data["position"]

            rect = Rect(paddle_position[0], paddle_position[1], paddle_width, paddle_height)
            x_coordinates = [rect.topleft[0], rect.topright[0], rect.bottomright[0], rect.bottomleft[0]]
            y_coordinates = [rect.topleft[1], rect.topright[1], rect.bottomright[1], rect.bottomleft[1]]
            polygon = draw_custom_polygon(
                x_coordinates=x_coordinates,
                y_coordinates=y_coordinates,
                color=self.config.platform_color_map[agent_name] if self.config.platform_color_map else "black",
                name=f"{agent_name} area",
            )
            polygons.append(polygon)

        data = self.config.episode_artifact.steps[step_idx].platforms[0]
        ball_measurement = data["sensors"]["Ball_Sensor"]["measurement"].m
        ball_radius = Ball().radius
        ball_x_coordinates = []
        ball_y_coordinates = []
        ball_x = ball_measurement[0]
        ball_y = ball_measurement[1]
        t = 0.0
        while t <= 2 * np.pi:
            ball_x_coordinates.append(ball_radius * np.cos(t) + ball_x)
            ball_y_coordinates.append(ball_radius * np.sin(t) + ball_y)
            t += 0.1

        ball = draw_custom_polygon(x_coordinates=ball_x_coordinates, y_coordinates=ball_y_coordinates, name="ball")
        polygons.append(ball)

        return polygons



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from __future__ import annotations

from pydantic import BaseModel

from .alerts import AlertGenerator


class ScenarioAlertGenerators(BaseModel):
    """Set of metric generators for a scenario"""

    world: list[AlertGenerator]
    default_participant: list[AlertGenerator]
    specific_participant: dict[str, list[AlertGenerator]]
    raise_on_error: bool = True

    @staticmethod
    def from_dict(data: dict, raise_on_error: bool = True) -> ScenarioAlertGenerators:
        """Create an instance from a dictionary"""

        if "world" in data:
            world_alert_generator = [AlertGenerator(**alert) for alert in data["world"]]
        default_agent_alert_generators: list[AlertGenerator] = []
        agent_alert_generators: dict[str, list[AlertGenerator]] = {}
        if "agent" in data:
            for key in data["agent"]:
                if key == "__default__":
                    default_agent_alert_generators.extend([AlertGenerator(**alert) for alert in data["agent"]["__default__"]])
                else:
                    agent_alert_generators[key].extend([AlertGenerator(**alert) for alert in data["agent"][key]])

        return ScenarioAlertGenerators(
            world=world_alert_generator,
            default_participant=default_agent_alert_generators,
            specific_participant=agent_alert_generators,
            raise_on_error=raise_on_error,
        )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import logging

import numpy as np

from corl.libraries.utils import get_wrap_diff
from corl.rewards.base_measurement_operation import BaseMeasurementOperation, BaseMeasurementOperationValidator


class ExponentialDecayFromTargetValueValidator(BaseMeasurementOperationValidator):
    """
    reward scale: scale of this reward, this would be the maximum reward value for
                  a given time step
    eps: the length of the reward curve for the exponential decay, would recommend playing with
        this value in a plotting software to determine the value you need
    target_value: the value with which to take the difference from the wrapped observation value
    index: the index with which to pull data out of the observation extractor, useful if len(observation) > 1
    is_wrap: if the obs difference needs to wrap around 0/360
    is_rad: if the obs difference is in terms of radians
    closer: to only reward with this reward if the difference to the target value
            is less than the last time step
    closer_tolerance: the difference tolerance at which point the agent is close enough
                      to the target value that "closer" is not a concern
    """

    reward_scale: float
    eps: float
    target_value: float = 0
    index: int | None = 0
    is_wrap: bool = False
    is_rad: bool = False
    method: bool = False
    closer: bool = False
    closer_tolerance: float = 0.0


class ExponentialDecayFromTargetValue(BaseMeasurementOperation):
    """
    Exponential Decay from Target Value
    wraps some sort of observation and takes in a target value,
    the reward based on the difference between the target value
    and the observation. the reward exponentially decays in value
    the further you are away from the target value. causing a spike at the target value
    """

    @staticmethod
    def get_validator() -> type[ExponentialDecayFromTargetValueValidator]:
        return ExponentialDecayFromTargetValueValidator

    def __init__(self, **kwargs) -> None:
        self.config: ExponentialDecayFromTargetValueValidator
        super().__init__(**kwargs)
        self._last_value = None
        self._logger = logging.getLogger(self.name)

    def __call__(
        self,
        observation,
        action,
        next_observation,
        state,
        next_state,
        observation_space,
        observation_units,
    ) -> float:
        reward = 0.0
        if self.config.agent_name not in next_observation:
            return reward

        obs = self.extractor.value(next_observation[self.config.agent_name])[self.config.index].m

        if self.config.is_wrap:
            # Note: this always returns the min angle diff and is positive
            diff = get_wrap_diff(obs, self.config.target_value, self.config.is_rad, self.config.method)
        else:
            diff = obs - self.config.target_value

        abs_diff = abs(diff)
        if self._last_value is None:
            self._last_value = abs_diff  # type: ignore

        func_applied: float = 0.0
        if not self.config.closer or ((self._last_value >= abs_diff) or abs_diff < self.config.closer_tolerance):  # type: ignore
            func_applied = np.exp(-np.abs(diff / self.config.eps))

        reward = self.config.reward_scale * func_applied

        self._last_value = abs_diff  # type: ignore
        return reward



"""

Confi File Watcher Module
"""
import logging
import logging.config
import os
import pathlib
import pprint
from datetime import datetime
from logging import FileHandler
from logging.handlers import RotatingFileHandler

import yaml
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

FILE_PATH = str(pathlib.Path(__file__).parent.absolute())


class PrettyLog:
    """Print util for dicts using logger"""

    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return pprint.pformat(self.obj)


class PrettyLogNL:
    """Print util for dicts using logger"""

    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return "\n" + pprint.pformat(self.obj)


class LoggingSetup:
    """
    Logging Setup Config
    """

    def __init__(
        self,
        default_path: str = os.path.join(os.getcwd(), "logging.yml"),
        default_level=logging.INFO,
        env_key: str = "LOG_CFG",
        enabled_file_watch=False,
    ):
        if enabled_file_watch:
            self._create_event_handler()
            self._setup_event_handler()
            self._create_observers()
        self._default_path = default_path
        self._default_level = default_level
        self._env_key = env_key
        self._setup_logging(self._default_path, self._default_level)
        self._log = logging.getLogger(LoggingSetup.__name__)

    def _setup_logging(self, default_path, default_level):
        self._log_file = default_path
        if os.path.exists(self._log_file):
            with open(self._log_file, encoding="utf8") as f:
                config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)
        else:
            logging.basicConfig(level=default_level)

    def _create_event_handler(self):
        self._patterns = "*"
        self._ignore_patterns = ""
        self._ignore_directories = False
        self._case_sensitive = True
        self._event_handler = PatternMatchingEventHandler(
            self._patterns,
            self._ignore_patterns,
            self._ignore_directories,
            self._case_sensitive,
        )

    def _setup_event_handler(self):
        self._event_handler.on_created = self._on_created
        self._event_handler.on_deleted = self._on_deleted
        self._event_handler.on_modified = self._on_modified
        self._event_handler.on_moved = self._on_moved

    def _create_observers(self):
        self._path = "."
        self._go_recursively = True
        self._observer = Observer()
        self._observer.schedule(self._event_handler, self._path, recursive=self._go_recursively)
        self._observer.start()

    def _on_created(self, event):
        self._log.debug(f"{event.src_path} has been created!")

    def _on_deleted(self, event):
        self._log.debug(f"{event.src_path} has been deleted!")

    def _on_modified(self, event):
        self._log.debug(f"{event.src_path} has been modified!")
        if self._log_file in event.src_path:
            self._setup_logging(self._default_path, self._default_level, self._env_key)

    def _on_moved(self, event):
        self._log.debug(f"{event.src_path} has been moved!")


class TimestampedFileHandler(FileHandler):
    """FileHandler that adds timestamp to log filename"""

    def __init__(self, filename, mode="a", encoding=None, delay=False):
        FileHandler.__init__(
            self,
            TimestampedFileHandler.timestamp_filename(filename),
            mode,
            encoding,
            delay,
        )

    @staticmethod
    def timestamp_filename(base_filename: str) -> str:
        """insert timestamp into filename as a prefix

        Parameters
        ----------
        base_filename : str
            the filename to insert timestamp into

        Returns
        -------
        str
            the filename with timestamp inserted
        """
        path, filename = os.path.split(base_filename)
        t = datetime.utcnow().strftime("%Y-%m-%d_%H-%M-%S.%fZ")
        filename = f"{t}_{filename}"
        return os.path.join(path, filename)


class TimestampedRotatingFileHandler(RotatingFileHandler):
    """RotatingFileHandler that adds timestamp to log filename"""

    def __init__(self, filename, mode="a", maxBytes=0, backupCount=0, encoding=None, delay=False):
        RotatingFileHandler.__init__(
            self,
            TimestampedFileHandler.timestamp_filename(filename),
            mode,
            maxBytes,
            backupCount,
            encoding,
            delay,
        )
        self._custom_base_filename = filename

    @staticmethod
    def _rotation_filename(default_name, postfix=""):
        filename = TimestampedFileHandler.timestamp_filename(default_name)
        filename, ext = os.path.splitext(filename)
        return filename + postfix + ext

    def doRollover(self, postfix=""):
        """
        do a rollover; in this case, a date/time stamp is appended to the filename
        when the rollover happens.  However, you want the file to be named for the
        start of the interval, not the current time.  If there is a backup count,
        then we have to get a list of matching filenames, sort them and remove
        the one with the oldest suffix.
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        self.baseFilename = self._rotation_filename(self._custom_base_filename, postfix)
        if not self.delay:
            self.stream = self._open()
